#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Dec 16 14:27:24 2019

@author: maksbh
"""


import numpy as np

"""
Function to find the nearest index 

Parameters
----------

array : Numpy array
value : value of energy

Returns
-------
idx : Integer
      index location corresponding to the closest location
"""

def find_nearest(array, value):
    idx = (np.abs(array - value)).argmin()
    return idx

"""
Function to get the interpolated value

Parameters
----------

array : Numpy array
value : value of energy
nearest_id : id corresponding to the nearest value

Returns
-------
valArray : Numpy array
           array of the interpolated values
"""
def get_interpolated_value(array,value,nearest_id,energy_id):
    valArray = np.zeros(array.shape[1]);
    if(array[nearest_id][energy_id] > value):
        xp = [array[nearest_id][energy_id], array[nearest_id - 1][energy_id]];
        for i in range(0,energy_id):
            yp = [array[nearest_id][i], array[nearest_id - 1][i]];
            valArray[i] = np.interp(value,xp,yp);
            
    elif (array[nearest_id][energy_id] < value):
        xp = [array[nearest_id][energy_id], array[nearest_id + 1][energy_id]];
        for i in range(0,energy_id):
            yp = [array[nearest_id][i], array[nearest_id + 1][i]];
            valArray[i] = np.interp(value,xp,yp);
            
    else:
        for i in range(0,len(valArray)):
            valArray[i] = value[nearest_id][i];
            
    valArray[energy_id] = value;
    return valArray;
         
def dump_dataVacuum(index,energy,f):
    Header = "EnergyData" + str(index) +":\n{\n";
    f.write(Header);
    Energy = "Energy = " + str(energy) + ";\n";
    f.write(Energy);
    BetaPara = "BetaPara = " + str(0.0) + ";\n";
    f.write(BetaPara);
    BetaPerp = "BetaPerp = " + str(0.0) + ";\n";
    f.write(BetaPerp);
    DeltaPara = "DeltaPara = " + str(0.0) + ";\n";
    f.write(DeltaPara);
    DeltaPerp = "DeltaPerp = " + str(0.0) + ";\n";
    f.write(DeltaPerp);
    f.write("}\n");
       
    
def dump_data(valArray,index,labelEnergy,f):
    Header = "EnergyData" + str(index) +":\n{\n";
    f.write(Header);
    Energy = "Energy = " + str(valArray[labelEnergy["Energy"]]) + ";\n";
    f.write(Energy);
    BetaPara = "BetaPara = " + str(valArray[labelEnergy["BetaPara"]]) + ";\n";
    f.write(BetaPara);
    BetaPerp = "BetaPerp = " + str(valArray[labelEnergy["BetaPerp"]]) + ";\n";
    f.write(BetaPerp);
    DeltaPara = "DeltaPara = " + str(valArray[labelEnergy["DeltaPara"]]) + ";\n";
    f.write(DeltaPara);
    DeltaPerp = "DeltaPerp = " + str(valArray[labelEnergy["DeltaPerp"]]) + ";\n";
    f.write(DeltaPerp);
    f.write("}\n");
     

    

def write_materials(startEnergy, endEnergy,increment,dict,labelEnergy,numMaterial):
    NumEnergy = int(np.round((endEnergy - startEnergy)/increment + 1));
    
    for numMat in range(0,numMaterial):
        f = open("Material" + str(numMat+1) + ".txt", "w")
        fname = dict["Material" + str(numMat)]
        if(fname != 'vacuum'):
            Data = np.loadtxt(fname,skiprows=1);
            Data = Data[Data[:,labelEnergy["Energy"]].argsort()]
            for i in range(0,NumEnergy):
                currentEnergy = startEnergy + i* increment;
                nearest_id = find_nearest(Data[:,labelEnergy["Energy"]],currentEnergy)
                ValArray = get_interpolated_value(Data,currentEnergy,nearest_id,labelEnergy["Energy"])
                dump_data(ValArray,i,labelEnergy,f)   
            
        else:
            for i in range(0,NumEnergy):
                energy = startEnergy + increment*i
                dump_dataVacuum(i,energy,f)
        f.close()    
    
    
    
if __name__ == "__main__":
    startEnergy = 280.0; #Start energy
    endEnergy = 290.0;   #End  energy
    incrementEnergy = 0.1; #Increment  energy
    startAngle = 0.0; #start angle
    endAngle = 180.0; #end angle
    incrementAngle = 2.0; #increment in each angle
    numThreads = 4; #number of threads for execution
    numX = 256; # number of voxels in X direction
    numY = 256;# number of voxels in Y direction
    numZ = 32;# number of voxels in Z direction
    PhysSize = 2.0;
    
    #Files corresponding to Each material. For vacuum pass null
    dict={'Material0':'PSrollinglog.txt',
          'Material1':'Au.txt',
          'Material2':'PEOlig2018.txt',
          'Material3':'vacuum'}
    
    
    # Label of energy to look for
    labelEnergy={"BetaPara":0,
                 "BetaPerp":1,
                 "DeltaPara":2,
                 "DeltaPerp":3,
                 "Energy":6}

#### Do not change below this
    f = open("config.txt", "w") 
    f.write("StartEnergy = " + str(startEnergy) + ";\n");
    f.write("EndEnergy = " + str(endEnergy) + ";\n");
    f.write("IncrementEnergy = " + str(incrementEnergy) + ";\n");
    f.write("StartAngle = " + str(startAngle) + ";\n");
    f.write("EndAngle = " + str(endAngle) + ";\n");
    f.write("IncrementAngle = " + str(incrementAngle) + ";\n");
    f.write("NumThreads = " + str(numThreads) + ";\n");
    f.write("NumX = " + str(numX) + ";\n");
    f.write("NumY = " + str(numY) + ";\n");
    f.write("NumZ = " + str(numZ) + ";\n");
    f.write("PhysSize = " + str(PhysSize) + ";\n");
    f.close();
    
    
    
    main(startEnergy,endEnergy,incrementEnergy,dict,labelEnergy,len(dict));



from lib.util import center_sphere, write_sphere_hdf5, write_config2D, write_slurm
from lib.generateConstants import *
from lib.reduce import *
import pathlib
import subprocess
import cupy as cp
import os
import h5py
import datetime
from numpy import genfromtxt
import shutil
from shutil import copyfile
import sys
import xarray as xr
import time
import matplotlib.pyplot as plt
plt.rcParams.update({'font.size':14})

#Generates Euler morphology from scratch
def coreshell_spheres_decay_euler(basePath, r=512,d=32,vx_size=2.5,ra=4,t=2.94,phi_iso=0.46,decay_order=0.42,numMaterial=3,
                                  filename="coreshelltest.hdf5",mformat='ZYX', 
                                  return_folderpath=False, hdfcompress=True, slurm_file=0, cyrsoxs_version='latest'):
    
    # a b c are boolean masks for material presence
    a_b = cp.full((d, r, r), False)  # core
    b_b = cp.full((d, r, r), False)  # shell
    c_b = cp.full((d, r, r), False)  # everything else

    #orientation components of shell
    b_x = cp.zeros([d, r, r])  # shell
    b_y = cp.zeros([d, r, r])  # shell
    b_z = cp.zeros([d, r, r])  # shell

    #volume fractions of each material - euler
    vf_a = cp.zeros([d, r, r])  # core
    vf_b = cp.zeros([d, r, r])  # shell
    vf_c = cp.zeros([d, r, r])  # everything else

    #euler angles and aligned fraction - euler morphology - ZXZ rotations
    SE_b = cp.zeros([d, r, r])
    theta_b = cp.zeros([d, r, r]) #theta of shell
    psi_b = cp.zeros([d, r, r]) #psi of shell

    #intermediate orientation components of shell
    b_xi = cp.zeros([d, r, r])  # shell
    b_yi = cp.zeros([d, r, r])  # shell
    b_zi = cp.zeros([d, r, r])  # shell

    #read list of coordinates
    l = genfromtxt('./lib/CoreShell/LoG_coord.csv', delimiter=',', skip_header=1)

    z,y,x = cp.ogrid[0:d:1,0:r:1, 0:r:1]

    for p in l:
        #print(str(p[0]) + " " + str(p[1]))
        # these are the particle coordinate locations
        #this mask defines each sphere
        #mf is a volume matrix (same shape as morphology) of the squared radius out from center
        mf = (x-p[0])**2 + (y-p[1])**2 +(z-15)**2 
        #create a boolean matrix (same shape as morphology) that is true if mf is less than squared radius
        mask = mf <= ra**2
        #add that sphere to material a boolean mask containing all the material a (core) spheres
        a_b = cp.logical_or(a_b, mask)
        #calculate the shell using the same mf matrix    
        mask = (a_b == False) & (mf <= (ra+t)**2)
        #add that shell to material b boolean mask containing all the material b shells
        b_b = cp.logical_or(b_b, mask)

        #these 3 statements calculate orientation vector pieces
        b_xi = ((x-p[0]) + y*0 + z*0)*mask
        b_yi = (x*0 + (y-p[1]) + z*0)*mask
        b_zi = (x*0 + y*0 + (z-15))*mask

        b_x = b_x + b_xi*(b_x==0)
        b_y = b_y + b_yi*(b_y==0)
        b_z = b_z + b_zi*(b_z==0)

    #ensure shell where there is no core and where there is shell
    b_b = (a_b == False) & (b_b == True)
    #orientation in shell only where there is shell
    b_x = b_x*b_b
    b_y = b_y*b_b
    b_z = b_z*b_b

    #matrix - 3rd component is everywhere that core and shell are not
    c_b = (a_b == False) & (b_b == False)

    #calculate the magnitude of the orientation vector in shell for later normalization
    b_t = (b_x**2 + b_y**2+b_z**2)**0.5

    #make each of the components of orientation normalized so that their squared sum is equal to aligned volume fraction
    #this part needs to be adjusted for decay; in principle we already have all the pieces... instead of dividing by b_t (hypotenuse length) we can add a ratio of (1st_shell/b_t)^decay_order, which will always be <=1  
    #what should 1st shell be? It needs to be a matrix of value ra
    b_1 = (ra)*b_b

    #total and aligned volume fractions
    vf_b = 1*b_b  #total volume fraction of shell = 1
    SE_b = b_b*(1-phi_iso)*(b_1/b_t)**(decay_order)  #aligned fraction of shell
    vf_a = 1*a_b  # total volume fraction of core = 1
    vf_c = 1*c_b  # total volume fraction of everything else = 1

    # convert the handful of NaN's arising from divide by zero to zeros
    SE_b = cp.nan_to_num(SE_b)

    # Euler angle calculations
    theta_b = cp.arctan2(((b_x)**2+(b_y)**2)**0.5, b_z)  #theta of shell - calculated from vector morphology
    psi_b = cp.arctan2(b_y, b_x)

    with h5py.File(filename, "w") as f:
        f.create_dataset('igor_parameters/igorefield', data="0,1")
        f.create_dataset('igor_parameters/igormaterialnum', data=numMaterial)
        f.create_dataset('igor_parameters/igornum', data=r)
        f.create_dataset('igor_parameters/igorthickness', data=d)
        f.create_dataset('igor_parameters/igorvoxelsize', data=vx_size)

        f.create_dataset('Morphology_Parameters/creation_date', data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S")) # datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
        f.create_dataset('Morphology_Parameters/film_normal', data=[1,0,0])
        f.create_dataset('Morphology_Parameters/morphology_creator', data="")
        f.create_dataset('Morphology_Parameters/name', data=filename)
        f.create_dataset('Morphology_Parameters/version', data="")#data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
        f.create_dataset('Morphology_Parameters/PhysSize', data=vx_size)
        f.create_dataset('Morphology_Parameters/NumMaterial',data=numMaterial)
        f.create_dataset('Morphology_Parameters/Parameters',data="r,d,ra,t,phi_iso,decay_order")
        f.create_dataset('Morphology_Parameters/Parameter_values',data=[r,d,ra,t,phi_iso,decay_order])


        mat1vf = f.create_dataset('Euler_Angles/Mat_1_Vfrac', data=cp.asnumpy(vf_a.astype(cp.float64)), compression="gzip", compression_opts=9)
        mat1s = f.create_dataset('Euler_Angles/Mat_1_S', data=np.zeros([d, r, r]), compression="gzip", compression_opts=9)
        mat1theta = f.create_dataset('Euler_Angles/Mat_1_Theta', data=np.zeros([d, r, r]), compression="gzip", compression_opts=9)
        mat1psi = f.create_dataset('Euler_Angles/Mat_1_Psi', data=np.zeros([d, r, r]), compression="gzip", compression_opts=9)
        mat2vf = f.create_dataset('Euler_Angles/Mat_2_Vfrac', data=cp.asnumpy(vf_b.astype(cp.float64)), compression="gzip", compression_opts=9)
        mat2s = f.create_dataset('Euler_Angles/Mat_2_S', data=cp.asnumpy(SE_b), compression="gzip", compression_opts=9)
        mat2theta = f.create_dataset('Euler_Angles/Mat_2_Theta', data=cp.asnumpy(theta_b), compression="gzip", compression_opts=9)
        mat2psi = f.create_dataset('Euler_Angles/Mat_2_Psi', data=cp.asnumpy(psi_b), compression="gzip", compression_opts=9)      
        mat3vf = f.create_dataset('Euler_Angles/Mat_3_Vfrac', data=cp.asnumpy(vf_c.astype(cp.float64)), compression="gzip", compression_opts=9)
        mat3s = f.create_dataset('Euler_Angles/Mat_3_S', data=np.zeros([d, r, r]), compression="gzip", compression_opts=9)
        mat3theta = f.create_dataset('Euler_Angles/Mat_3_Theta', data=np.zeros([d, r, r]), compression="gzip", compression_opts=9)     
        mat3psi = f.create_dataset('Euler_Angles/Mat_3_Psi', data=np.zeros([d, r, r]), compression="gzip", compression_opts=9)

        for i in range(1,numMaterial+1):
            for j in range(0,3):
                locals()[f"mat{i}vf"].dims[j].label = mformat[j]
                locals()[f"mat{i}s"].dims[j].label = mformat[j]
                locals()[f"mat{i}theta"].dims[j].label = mformat[j]
                locals()[f"mat{i}psi"].dims[j].label = mformat[j]
    f.close()
    
    # Create folder stamped with current datetime
    time_now = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    folder_name = 'CoreShelltest_' + time_now
    os.mkdir(folder_name)
    subprocess.run(["cp","./lib/CoreShell/Material1.txt",f"./{folder_name}"])
    subprocess.run(["cp","./lib/CoreShell/Material2.txt",f"./{folder_name}"])
    subprocess.run(["cp","./lib/CoreShell/Material3.txt",f"./{folder_name}"])
    subprocess.run(["cp","./lib/CoreShell/config.txt",f"./{folder_name}"])
    subprocess.run(["mv",f"{filename}",f"./{folder_name}"])

    if slurm_file == 1:
        write_slurm(filename, cyrsoxs_version=cyrsoxs_version)
        subprocess.run(["mv","job.slurm",f"./{folder_name}"])
        
    
    if return_folderpath:
        return pathlib.Path(basePath,folder_name)

    
    
def run_slurm(folder_path):
    os.chdir(folder_path)
    result = subprocess.run(['sbatch','job.slurm'])
    return result


def test_vs_reference(folder_path,cyrsoxs_version='latest', savepng=False):
    h5_path = pathlib.Path(folder_path,'HDF5')
    h5list = sorted(list(h5_path.glob('*h5')))
    while len(h5list) < 101:
        time.sleep(0.5)
        h5list = sorted(list(h5_path.glob('*h5')))
        
    horz = np.zeros((len(h5list),363))
    vert = horz.copy()
    for i, file in enumerate(h5list):
        data = read_img_h5(file,PhysSize=2.5)
        remeshed = remesh(data)
        horz[i,:] = remeshed.sel(chi=slice(170,190)).mean('chi')
        vert[i,:] = remeshed.sel(chi=slice(80,100)).mean('chi')
        
    with np.errstate(invalid='ignore'):
        A_np = -(vert-horz)/(vert+horz)
    
    A = xr.DataArray(A_np, dims=['energy','q'],coords={'q':remeshed.q,'energy':np.round(np.arange(280.0,290.01,0.1),1)}).sortby('energy')
    
    A_reference = xr.open_dataarray('/home/pjd1/CyRSoXS-Processing/pete/test_directory/lib/CoreShell/CS_reference.nc')
    A_reference = A_reference.interp(energy=np.round(np.arange(280.0,290.01,0.1),1)).sortby('energy')
    if savepng:
        plt.figure()
#         (A_reference.sel(q=slice(0.02,0.4)).mean('q') - A.sel(q=slice(0.02,0.4)).mean('q')).plot(color='k',label='Reference',xscale='log')
        A_reference.sel(q=slice(0.02,0.4)).mean('q').plot(color='k',label='Reference',xscale='log')
        A.sel(q=slice(0.02,0.4)).mean('q').plot(linestyle='--',label='CyRSoXS',xscale='log')
        plt.legend()
        plt.xlabel('Energy [eV]')
        plt.ylabel('A(E)')
        plt.title('')
        plt.xticks(ticks=[280,282,284,286,288,290],labels=[280,282,284,286,288,290])
        plt.savefig(pathlib.Path(folder_path,'AvE.png'),format='png',dpi=140,bbox_inches='tight')
        plt.close()
        
        plt.figure()
#         (A_reference.sel(energy=284.7,q=slice(0.02,0.4)) - A.sel(energy=284.7,q=slice(0.02,0.4))).plot(xscale='log',color='k',label=f'284.7 eV Reference')
#         (A_reference.sel(energy=285.2,q=slice(0.02,0.4)) -  A.sel(energy=285.2,q=slice(0.02,0.4))).plot(xscale='log',color='r',label=f'285.2 eV Reference')
        A_reference.sel(energy=284.7,q=slice(0.02,0.4)).plot(xscale='log',color='k',label=f'284.7 eV Reference')
        A_reference.sel(energy=285.2,q=slice(0.02,0.4)).plot(xscale='log',color='r',label=f'285.2 eV Reference')
        A.sel(energy=284.7,q=slice(0.02,0.4)).plot(xscale='log',linestyle='--',label=f'284.7 eV CyRSoXS')
        A.sel(energy=285.2,q=slice(0.02,0.4)).plot(xscale='log',linestyle='--',label=f'285.2 eV CyRSoXS')
        plt.xlim(0.02,0.41)
        plt.ylim(-0.3,0.3)
        plt.legend()
        plt.xlabel(r'q [nm$^{-1}$]')
        plt.ylabel('A(q)')
        plt.title('')
        plt.savefig(pathlib.Path(folder_path,'AvQ.png'),format='png',dpi=140,bbox_inches='tight')
        plt.close()

    
    assert(np.allclose(A_reference.sel(q=slice(0.02,0.4)).mean('q').values,A.sel(q=slice(0.02,0.4)).mean('q').values,atol=2e-6,equal_nan=True))
    assert(np.allclose(A_reference.sel(energy=284.7,q=slice(0.02,0.4)).values,A.sel(energy=284.7,q=slice(0.02,0.4)).values,atol=1e-4,equal_nan=True))
    assert(np.allclose(A_reference.sel(energy=285.2,q=slice(0.02,0.4)).values,A.sel(energy=285.2,q=slice(0.02,0.4)).values,atol=1e-4,equal_nan=True))
    
if __name__ == '__main__':
    version = 'latest'
    slurm_file = 0
    
    for arg in sys.argv:
        if arg.startswith('slurm'):
            slurm_file = int(arg.split('=')[-1])
        if arg.startswith('version'):
            version = arg.split('=')[-1]

    basePath = pathlib.Path('.').absolute()
    
    workpath = coreshell_spheres_decay_euler(basePath, filename="coreshelltest.hdf5",mformat='ZYX',return_folderpath=True, hdfcompress=True, slurm_file=slurm_file, cyrsoxs_version=version)

    if slurm_file == 1:
        import time
        result = run_slurm(workpath)
        if result.returncode == 0:
            test_vs_reference(workpath,cyrsoxs_version=version, savepng=True)


import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib import rc, gridspec
import matplotlib
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import numpy as np
import warnings
import pathlib


def morphology_visualizer(
    morphology,
    z_slice: int = 0,
    subsample: int = None,
    translate_x: int = None,
    translate_y: int = None,
    screen_euler: bool = True,
    add_quiver: bool = False,
    quiver_bw: bool = True,
    outputmat: list = None,
    outputplot: list = None,
    outputaxes: bool = True,
    vfrac_range: list = None,
    S_range: list = None,
    vfrac_cmap: str = None,
    S_cmap: str = None,
    runquiet: bool = False,
    batchMode: bool = False,
    plotstyle: str = "light",
    dpi: int = 300,
    exportDir: str = None,
    exportParams: dict = None,
):
    """
    Reads in morphology HDF5 file and checks that the format is consistent for CyRSoXS. Optionally plots and returns select quantities.

    Parameters
    ----------

        z_slice : int
            Which z-slice of the array to plot.
        subsample : int
            Number of voxels to display in X and Y
        translate_x : int
            Number of voxels to translate image in x; meant for use with subsample
        translate_y : int
            Number of voxels to translate image in y; meant for use with subsample
        screen_euler : bool
            Suppress visualization of euler angles where vfrac < 0.05 or S < 0.05; intended to hilight edges
        add_quiver : bool
            Adds lines to every voxel on the psi plot that indicate in-plane direction. Not recommended for resolutions larger than 128x128, best for resolutions 64x64 or lower.
        quiver_bw : bool
            Intended to be used when add_quiver == True, when quiver_bw is True, the quiver arrows will be black and white instead of colored.
        outputmat : list of ints
            Number of which materials to return
        outputplot : list of strings
            Number of which plots to return, can include 'vfrac', 'S', 'theta', 'psi'
        outputaxes : bool
            If a plot is returned, include its axes
        vfrac_range: list of tuples as [float, float]
            A custom range for vfrac colorbar
        S_range: list of tuples as [float, float]
            A custom range for S colorbar
        vfrac_cmap: str
            A custom substitution for vfrac colormap
        S_cmap: str
            A custom substitution for vfrac colormap
        runquiet : bool
            Boolean flag for running without plotting or outputting to console
        batchMode : bool
            if true, prints console output and generates plots but doesnt show (provide exportDir for export)
        plotstyle : str
            Use a light or dark background for plots. 'dark' - dark, 'light' - light
        dpi : int
            The dpi at which the plot is generated. Per-material plot dimensions are 8.5" x 12.75"
        exportDir : str, optional
            if provided, export directory to save any generated figures into,
            by default, will respect dpi and save as png, use exportParams to override
        exportParams : dict, optional
            additional params to unpack into matplotlib.pyplot.savefig. Overrides existing params.
            ex: exportParams = {'dpi':600, format='svg'}
    Returns
    -------
        If outputmat and outputplot are correctly entered, will return an index list of images of the selected material and plot. Each list element will be  a numpy array in RGB format that be displayed with imshow

    """
    style_dict = {"dark": "dark_background", "light": "default"}

    rgb_return_list = []

    if subsample is None:
        subsample = morphology.NumZYX[1]  # y dimension

    if not runquiet:
        print(
            f"Dataset dimensions (Z, Y, X): {morphology.NumZYX[0]} x {morphology.NumZYX[1]} x"
            f" {morphology.NumZYX[2]}"
        )
        print(f"Number of Materials: {morphology._numMaterial}")
        print("")

    if z_slice > (morphology.NumZYX[0] - 1):
        warnings.warn(
            f"z_slice of {z_slice} is greater than the maximum index of {morphology.NumZYX[0]-1}."
            f" Using z_slice = {morphology.NumZYX[0]-1} instead."
        )
        z_slice = morphology.NumZYX[0] - 1

    #         if plotstyle == 'dark':
    plt.style.use(style_dict[plotstyle])
    font = {
        "family": "sans-serif",
        "sans-serif": "DejaVu Sans",
        "weight": "regular",
        "size": 8,
    }

    rc("font", **font)

    cwdPath = pathlib.Path(__file__).resolve().parent
    psi_cmap = matplotlib.colors.ListedColormap(np.load(cwdPath / "cmap/infinitydouble_cmap.npy"))

    backend_ = matplotlib.get_backend()

    try:
        if runquiet:
            matplotlib.use("Agg")  # Prevent showing stuff

        for i in range(1, morphology._numMaterial + 1):
            fig = plt.figure(figsize=(8.5, 12.75), dpi=dpi)
            if runquiet == False:
                print(
                    f"Material {i} Vfrac. Min: {morphology.materials[i].Vfrac.min()} Max:"
                    f" {morphology.materials[i].Vfrac.max()}"
                )
                print(
                    f"Material {i} S. Min: {morphology.materials[i].S.min()} Max:"
                    f" {morphology.materials[i].S.max()}"
                )
                print(
                    f"Material {i} theta. Min: {morphology.materials[i].theta.min()} Max:"
                    f" {morphology.materials[i].theta.max()}"
                )
                print(
                    f"Material {i} psi. Min: {morphology.materials[i].psi.min()} Max:"
                    f" {morphology.materials[i].psi.max()}"
                )

            if (morphology.materials[i].theta.min() < 0) or (
                morphology.materials[i].theta.max() > (np.pi)
            ):
                warnings.warn(
                    "Visualization expects theta to have bounds of [0,pi]. This model has theta"
                    " outside those bounds and visualization may be incorrect."
                )

            # run if you don't want runquiet or run if you've selected this material for output
            if (runquiet is not True) or ((outputmat is not None) and (i in outputmat)):
                gs = gridspec.GridSpec(
                    nrows=5,
                    ncols=2,
                    figure=fig,
                    width_ratios=[1, 1],
                    height_ratios=[3, 1, 0.1, 3, 1],
                    wspace=0.3,
                    hspace=0.65,
                )

                start = int(morphology.NumZYX[1] / 2) - int(subsample / 2)
                end = int(morphology.NumZYX[1] / 2) + int(subsample / 2)

                if translate_x:
                    start_x = max(0, start + translate_x)
                    end_x = min(end + translate_x, morphology.NumZYX[1])
                else:
                    start_x = start
                    end_x = end

                if translate_y:
                    start_y = max(0, start + translate_y)
                    end_y = min(end + translate_y, morphology.NumZYX[1])
                else:
                    start_y = start
                    end_y = end

                if (runquiet is not True) or ("vfrac" in outputplot):
                    ax1 = plt.subplot(gs[0, 0])
                    if vfrac_range and (len(vfrac_range) >= i) and (len(vfrac_range[i - 1]) == 2):
                        norm = matplotlib.colors.Normalize(
                            vmin=vfrac_range[i - 1][0], vmax=vfrac_range[i - 1][1]
                        )
                    else:
                        norm = "linear"

                    if vfrac_cmap:
                        cmap = plt.get_cmap(vfrac_cmap)
                    else:
                        cmap = plt.get_cmap("winter")

                    Vfracplot = ax1.imshow(
                        morphology.materials[i].Vfrac[z_slice, :, :],
                        cmap=cmap,
                        origin="lower",
                        interpolation="none",
                        norm=norm,
                    )
                    ax1.set_ylabel("Y index", labelpad=0)
                    ax1.set_xlabel("X index")
                    ax1.set_title(f"Mat {i} {morphology.materials[i].name} Vfrac")
                    ax1.set_xlim(start_x, end_x)
                    ax1.set_ylim(start_y, end_y)
                    Vfrac_cbar = plt.colorbar(Vfracplot, ax=ax1, fraction=0.040)
                    # Vfrac_cbar.set_label(
                    #     "Vfrac: volume fraction", rotation=270, labelpad=22
                    # )

                if (runquiet is not True) or ("S" in outputplot):
                    ax2 = plt.subplot(gs[0, 1])
                    if S_range and (len(S_range) >= i) and (len(S_range[i - 1]) == 2):
                        norm = matplotlib.colors.Normalize(
                            vmin=S_range[i - 1][0], vmax=S_range[i - 1][1]
                        )
                    else:
                        norm = "linear"

                    if S_cmap:
                        cmap = plt.get_cmap(S_cmap)
                    else:
                        cmap = plt.get_cmap("nipy_spectral")

                    Splot = ax2.imshow(
                        morphology.materials[i].S[z_slice, :, :],
                        cmap=cmap,
                        origin="lower",
                        interpolation="none",
                        norm=norm,
                    )
                    ax2.set_ylabel("Y index", labelpad=0)
                    ax2.set_xlabel("X index")
                    ax2.set_title(f"Mat {i} {morphology.materials[i].name} S")
                    ax2.set_xlim(start_x, end_x)
                    ax2.set_ylim(start_y, end_y)
                    S_cbar = plt.colorbar(Splot, fraction=0.040)
                    # S_cbar.set_label(
                    #     "S: orientational order parameter", rotation=270, labelpad=22
                    # )

                # only do this if not runquiet; these plots are not outputted
                if runquiet is not True:
                    ax3 = plt.subplot(gs[1, 0])
                    ax3.hist(morphology.materials[i].Vfrac.flatten())
                    ax3.set_title(f"Mat {i} {morphology.materials[i].name} Vfrac")
                    ax3.set_xlim(left=0)
                    ax3.set_xlabel("Vfrac: volume fraction")
                    ax3.set_ylabel("num voxels")
                    ax3.set_yscale("log")

                    ax4 = plt.subplot(gs[1, 1])
                    ax4.hist(morphology.materials[i].S.flatten())
                    ax4.set_title(f"Mat {i} {morphology.materials[i].name} S")
                    ax4.set_xlim(left=0)
                    ax4.set_xlabel("S: orientational order parameter")
                    ax4.set_ylabel("num voxels")
                    ax4.set_yscale("log")

                if (runquiet is not True) or ("theta" in outputplot):
                    ax5 = plt.subplot(gs[3, 0])
                    norm = matplotlib.colors.Normalize(vmin=0, vmax=np.pi, clip=False)
                    if screen_euler:
                        thetaplot = ax5.imshow(
                            np.ma.masked_array(
                                morphology.materials[i].theta[z_slice, :, :] % np.pi,
                                np.logical_or(
                                    morphology.materials[i].Vfrac[z_slice, :, :] < 0.01,
                                    morphology.materials[i].S[z_slice, :, :] < 0.01,
                                ),
                            ),
                            cmap=plt.get_cmap("jet"),
                            norm=norm,
                            origin="lower",
                            interpolation="none",
                        )

                    else:
                        thetaplot = ax5.imshow(
                            morphology.materials[i].theta[z_slice, :, :] % np.pi,
                            cmap=plt.get_cmap("jet"),
                            norm=norm,
                            origin="lower",
                            interpolation="none",
                        )
                    ax5.set_ylabel("Y index")
                    ax5.set_xlabel("X index")
                    ax5.set_title(f"Mat {i} {morphology.materials[i].name} theta")
                    ax5.set_xlim(start_x, end_x)
                    ax5.set_ylim(start_y, end_y)
                    theta_cbar = plt.colorbar(thetaplot, fraction=0.040)
                    # theta_cbar.set_label("theta in radians", rotation=270, labelpad=22)

                    ax5i = inset_axes(
                        ax5,
                        axes_class=matplotlib.projections.get_projection_class("polar"),
                        width=0.7,
                        height=0.7,
                        axes_kwargs={"alpha": 0},
                    )

                    ax5i.grid(False)
                    # creates inset legend for orientation
                    azimuths_t = np.deg2rad(np.arange(-90, 90, 1))
                    zeniths_t = np.linspace(5, 10, 50)
                    values_t = np.mod(np.pi / 2 - azimuths_t, np.pi) * np.ones((50, 180))
                    ax5i.plot(
                        0,
                        0,
                        "o",
                        ms=75,
                        mec="none",
                        mfc=plt.rcParams["axes.facecolor"],
                        mew=2,
                        zorder=0,
                        alpha=0.7,
                    )
                    ax5i.pcolormesh(azimuths_t, zeniths_t, values_t, cmap=cm.jet, shading="auto")
                    ax5i.set_axis_off()
                    ax5i.arrow(
                        0, 0, 0, 4, width=0.005, head_width=0.2, head_length=0.6, lw=0.5
                    )  # ,facecolor='k',edgecolor='k')
                    ax5i.arrow(
                        np.pi / 2,
                        0,
                        0,
                        4,
                        width=0.005,
                        head_width=0.2,
                        head_length=0.6,
                        lw=0.5,
                    )  # , facecolor='k',edgecolor='k')
                    ax5i.text(2 * np.pi - np.deg2rad(40), 3.5, "X", fontsize=6)
                    ax5i.text(np.pi / 2 + np.deg2rad(40), 3.5, "Z", fontsize=6)

                if (runquiet is not True) or ("psi" in outputplot):
                    ax6 = plt.subplot(gs[3, 1])
                    norm = matplotlib.colors.Normalize(vmin=0, vmax=2 * np.pi, clip=False)
                    if screen_euler:
                        screen_mask = np.logical_or(
                            morphology.materials[i].Vfrac[z_slice, :, :] < 0.01,
                            morphology.materials[i].S[z_slice, :, :] < 0.01,
                        )
                        psiplot = ax6.imshow(
                            np.ma.masked_array(
                                morphology.materials[i].psi[z_slice, :, :] % (2 * np.pi),
                                screen_mask,
                            ),
                            cmap=psi_cmap,  # plt.get_cmap("hsv"),
                            norm=norm,
                            origin="lower",
                            interpolation="none",
                        )
                        if add_quiver:
                            screen_white = np.logical_or(
                                screen_mask,
                                (morphology.materials[i].psi[z_slice, :, :] % (2 * np.pi)) > np.pi,
                            )
                            screen_black = np.logical_or(
                                screen_mask,
                                (morphology.materials[i].psi[z_slice, :, :] % (2 * np.pi)) < np.pi,
                            )
                            sin_psi = np.sin(morphology.materials[i].psi[z_slice, :, :])
                            cos_psi = np.cos(morphology.materials[i].psi[z_slice, :, :])
                            len_scale = np.maximum(np.abs(sin_psi), np.abs(cos_psi))
                            if quiver_bw:
                                ax6.quiver(
                                    np.ma.masked_array(
                                        cos_psi / len_scale,
                                        screen_white,
                                    ),
                                    np.ma.masked_array(
                                        sin_psi / len_scale,
                                        screen_white,
                                    ),
                                    angles="xy",
                                    scale=1,
                                    pivot="mid",
                                    headaxislength=0,
                                    headlength=0,
                                    scale_units="xy",
                                    color="white",
                                )
                            else:
                                ax6.quiver(
                                    np.ma.masked_array(
                                        cos_psi / len_scale,
                                        screen_white,
                                    ),
                                    np.ma.masked_array(
                                        sin_psi / len_scale,
                                        screen_white,
                                    ),
                                    np.ma.masked_array(
                                        (morphology.materials[i].psi[z_slice, :, :] + np.pi)
                                        % (2 * np.pi),
                                        screen_white,
                                    ),
                                    cmap=psi_cmap,  # plt.get_cmap("hsv"),
                                    norm=norm,
                                    angles="xy",
                                    scale=1,
                                    pivot="mid",
                                    headaxislength=0,
                                    headlength=0,
                                    scale_units="xy",
                                )
                            if quiver_bw:
                                ax6.quiver(
                                    np.ma.masked_array(
                                        cos_psi / len_scale,
                                        screen_black,
                                    ),
                                    np.ma.masked_array(
                                        sin_psi / len_scale,
                                        screen_black,
                                    ),
                                    angles="xy",
                                    scale=1,
                                    pivot="mid",
                                    headaxislength=0,
                                    headlength=0,
                                    scale_units="xy",
                                    color="black",
                                )
                            else:
                                ax6.quiver(
                                    np.ma.masked_array(
                                        cos_psi / len_scale,
                                        screen_black,
                                    ),
                                    np.ma.masked_array(
                                        sin_psi / len_scale,
                                        screen_black,
                                    ),
                                    np.ma.masked_array(
                                        (morphology.materials[i].psi[z_slice, :, :] + np.pi)
                                        % (2 * np.pi),
                                        screen_black,
                                    ),
                                    cmap=psi_cmap,  # plt.get_cmap("hsv"),
                                    norm=norm,
                                    angles="xy",
                                    scale=1,
                                    pivot="mid",
                                    headaxislength=0,
                                    headlength=0,
                                    scale_units="xy",
                                )
                    else:
                        psiplot = ax6.imshow(
                            morphology.materials[i].psi[z_slice, :, :] % (2 * np.pi),
                            cmap=psi_cmap,  # plt.get_cmap("hsv"),
                            norm=norm,
                            origin="lower",
                            interpolation="none",
                        )
                        if add_quiver:
                            screen_white = (
                                morphology.materials[i].psi[z_slice, :, :] % (2 * np.pi) > np.pi
                            )
                            screen_black = (
                                morphology.materials[i].psi[z_slice, :, :] % (2 * np.pi) < np.pi
                            )
                            sin_psi = np.sin(morphology.materials[i].psi[z_slice, :, :])
                            cos_psi = np.cos(morphology.materials[i].psi[z_slice, :, :])
                            len_scale = np.maximum(np.abs(sin_psi), np.abs(cos_psi))
                            ax6.quiver(
                                np.ma.masked_array(
                                    cos_psi / len_scale,
                                    screen_white,
                                ),
                                np.ma.masked_array(
                                    sin_psi / len_scale,
                                    screen_white,
                                ),
                                angles="xy",
                                scale=1,
                                pivot="mid",
                                headaxislength=0,
                                headlength=0,
                                scale_units="xy",
                                color="white",
                            )
                            ax6.quiver(
                                np.ma.masked_array(
                                    cos_psi / len_scale,
                                    screen_black,
                                ),
                                np.ma.masked_array(
                                    sin_psi / len_scale,
                                    screen_black,
                                ),
                                angles="xy",
                                scale=1,
                                pivot="mid",
                                headaxislength=0,
                                headlength=0,
                                scale_units="xy",
                                color="black",
                            )

                    ax6.set_ylabel("Y index")
                    ax6.set_xlabel("X index")
                    ax6.set_title(f"Mat {i} {morphology.materials[i].name} psi")
                    ax6.set_xlim(start_x, end_x)
                    ax6.set_ylim(start_y, end_y)
                    psi_cbar = plt.colorbar(psiplot, fraction=0.040)
                    # psi_cbar.set_label("psi in radians", labelpad=22)
                    ax4i = inset_axes(
                        ax6,
                        axes_class=matplotlib.projections.get_projection_class("polar"),
                        width=0.7,
                        height=0.7,
                        axes_kwargs={"alpha": 0},
                    )

                    ax4i.grid(False)
                    # creates inset legend for orientation
                    azimuths = np.deg2rad(np.arange(0, 360, 1))
                    zeniths = np.linspace(5, 10, 50)
                    values = np.mod(azimuths, 2 * np.pi) * np.ones((50, 360))
                    ax4i.plot(
                        0,
                        0,
                        "o",
                        ms=75,
                        mec="none",
                        mfc=plt.rcParams["axes.facecolor"],
                        mew=2,
                        zorder=0,
                        alpha=0.7,
                    )
                    ax4i.pcolormesh(azimuths, zeniths, values, cmap=psi_cmap, shading="auto")
                    ax4i.set_axis_off()
                    ax4i.arrow(
                        0, 0, 0, 4, width=0.005, head_width=0.2, head_length=0.6, lw=0.5
                    )  # ,facecolor='k',edgecolor='k')
                    ax4i.arrow(
                        np.pi / 2,
                        0,
                        0,
                        4,
                        width=0.005,
                        head_width=0.2,
                        head_length=0.6,
                        lw=0.5,
                    )  # ,facecolor='k',edgecolor='k')
                    ax4i.text(2 * np.pi - np.deg2rad(40), 3.5, "X", fontsize=6)
                    ax4i.text(np.pi / 2 + np.deg2rad(40), 3.5, "Y", fontsize=6)

                if runquiet is not True:
                    ax7 = plt.subplot(gs[4, 0])
                    ax7.hist(morphology.materials[i].theta.flatten())
                    ax7.set_title(f"Mat {i} {morphology.materials[i].name} theta")
                    ax7.set_xlim(left=0)
                    ax7.set_xlabel("theta in radians")
                    ax7.set_ylabel("num voxels")
                    ax7.set_yscale("log")

                    ax8 = plt.subplot(gs[4, 1])
                    ax8.hist(morphology.materials[i].psi.flatten())
                    ax8.set_title(f"Mat {i} {morphology.materials[i].name} psi")
                    ax8.set_xlim(left=0)
                    ax8.set_xlabel("psi in radians")
                    ax8.set_ylabel("num voxels")
                    ax8.set_yscale("log")

            if runquiet is False:  # Show plot and/or export to file
                # Exporting plots
                if exportDir is not None:
                    # Attempt to export image
                    outDir = pathlib.Path(exportDir)
                    outPath = outDir / f"Mat{i}_viz"

                    if exportParams is None:  # No user provided kwargs
                        plt.savefig(fname=(str(outPath) + ".png"), dpi=dpi, format="png")
                    else:  # Apply user provided kwargs
                        # Grab format, if provided, else do png
                        format = exportParams.get("format", "png")
                        plt.savefig(fname=(str(outPath) + f".{format}"), **exportParams)
                if batchMode is True:  # Dont show plots, but do export them
                    pass
                else:
                    plt.show()

            if outputmat and (i in outputmat):
                fig.canvas.draw()
                rgb_return = np.array(fig.canvas.buffer_rgba(), dtype=np.uint8)
                rgb_return = rgb_return.reshape(fig.canvas.get_width_height()[::-1] + (4,))
                if outputplot and ("vfrac" in outputplot):
                    if outputaxes:
                        fig_xl = ax1.get_tightbbox().intervalx[0].astype(int)
                        fig_xr = Vfrac_cbar.ax.get_tightbbox().intervalx[1].astype(int)

                        fig_yt = rgb_return.shape[0] - ax1.get_tightbbox().intervaly[1].astype(int)
                        fig_yb = rgb_return.shape[0] - ax1.get_tightbbox().intervaly[0].astype(int)
                    else:
                        fig_yt = rgb_return.shape[0] - ax1.get_window_extent().intervaly[1].astype(
                            int
                        )
                        fig_yb = rgb_return.shape[0] - ax1.get_window_extent().intervaly[0].astype(
                            int
                        )
                        fig_xl = ax1.get_window_extent().intervalx[0].astype(int)
                        fig_xr = ax1.get_window_extent().intervalx[1].astype(int)

                    rgb_return_list.append(rgb_return[fig_yt:fig_yb, fig_xl:fig_xr])
                if outputplot and "S" in outputplot:
                    if outputaxes:
                        fig_yt = rgb_return.shape[0] - ax2.get_tightbbox().intervaly[1].astype(int)
                        fig_yb = rgb_return.shape[0] - ax2.get_tightbbox().intervaly[0].astype(int)
                        fig_xl = ax2.get_tightbbox().intervalx[0].astype(int)
                        fig_xr = S_cbar.ax.get_tightbbox().intervalx[1].astype(int)
                    else:
                        fig_yt = rgb_return.shape[0] - ax2.get_window_extent().intervaly[1].astype(
                            int
                        )
                        fig_yb = rgb_return.shape[0] - ax2.get_window_extent().intervaly[0].astype(
                            int
                        )
                        fig_xl = ax2.get_window_extent().intervalx[0].astype(int)
                        fig_xr = ax2.get_window_extent().intervalx[1].astype(int)
                    rgb_return_list.append(rgb_return[fig_yt:fig_yb, fig_xl:fig_xr])
                if outputplot and "theta" in outputplot:
                    if outputaxes:
                        fig_yt = rgb_return.shape[0] - ax5.get_tightbbox().intervaly[1].astype(int)
                        fig_yb = rgb_return.shape[0] - ax5.get_tightbbox().intervaly[0].astype(int)
                        fig_xl = ax5.get_tightbbox().intervalx[0].astype(int)
                        fig_xr = theta_cbar.ax.get_tightbbox().intervalx[1].astype(int)
                    else:
                        fig_yt = rgb_return.shape[0] - ax5.get_window_extent().intervaly[1].astype(
                            int
                        )
                        fig_yb = rgb_return.shape[0] - ax5.get_window_extent().intervaly[0].astype(
                            int
                        )
                        fig_xl = ax5.get_window_extent().intervalx[0].astype(int)
                        fig_xr = ax5.get_window_extent().intervalx[1].astype(int)
                    rgb_return_list.append(rgb_return[fig_yt:fig_yb, fig_xl:fig_xr])
                if outputplot and "psi" in outputplot:
                    if outputaxes:
                        fig_yt = rgb_return.shape[0] - ax6.get_tightbbox().intervaly[1].astype(int)
                        fig_yb = rgb_return.shape[0] - ax6.get_tightbbox().intervaly[0].astype(int)
                        fig_xl = ax6.get_tightbbox().intervalx[0].astype(int)
                        fig_xr = psi_cbar.ax.get_tightbbox().intervalx[1].astype(int)
                    else:
                        fig_yt = rgb_return.shape[0] - ax6.get_window_extent().intervaly[1].astype(
                            int
                        )
                        fig_yb = rgb_return.shape[0] - ax6.get_window_extent().intervaly[0].astype(
                            int
                        )
                        fig_xl = ax6.get_window_extent().intervalx[0].astype(int)
                        fig_xr = ax6.get_window_extent().intervalx[1].astype(int)
                    rgb_return_list.append(rgb_return[fig_yt:fig_yb, fig_xl:fig_xr])
        if not runquiet:
            plt.show()
        plt.clf()
        plt.close(fig)
    finally:
        # this code can hijack the backend and "agg" will not show figures, so this needs always to run, regardless of whether there are errors in the arguments.
        matplotlib.rc_file_defaults()
        matplotlib.use(backend_)
    return rgb_return_list



def read_material(filename):
    with open(filename, 'r') as f:
        lines = f.readlines()
    energy = []
    betapara = []
    betaperp = []
    deltapara = []
    deltaperp = []
    for line in lines:
        if line.startswith('Energy '):
            split_line = line.split(' = ')
            energy.append(float(split_line[1].strip(';\n')))
        elif line.startswith('BetaPara'):
            split_line = line.split(' = ')
            betapara.append(float(split_line[1].strip(';\n')))
        elif line.startswith('BetaPerp'):
            split_line = line.split(' = ')
            betaperp.append(float(split_line[1].strip(';\n')))
        elif line.startswith('DeltaPara'):
            split_line = line.split(' = ')
            deltapara.append(float(split_line[1].strip(';\n')))
        elif line.startswith('DeltaPerp'):
            split_line = line.split(' = ')
            deltaperp.append(float(split_line[1].strip(';\n')))

    deltabeta = [[val1, val2, val3, val4] for val1, val2, val3, val4 in zip(deltapara, betapara, deltaperp, betaperp)]
    deltabeta = dict(zip(energy, deltabeta))
    return energy, deltabeta


def read_config(config_file):
    config_dict = dict()
    with open(config_file, 'r') as f:
        d = f.readlines()

    # clean and split each line, adding to config dictionary
    for line in d:
        clean_line = line.strip(';\n')
        split_line = clean_line.split(' = ')
        # test if it's a list
        if '[' in split_line[1]:
            # if it is, parse string back to real list
            config_dict[split_line[0]] = [float(val) for val in split_line[1].strip('[]').split(',')]
        else:
            # not a list, assume bool or int for now
            if 'False' in split_line[1]:
                config_dict[split_line[0]] = False
            elif 'True' in split_line[1]:
                config_dict[split_line[0]] = True
            else:
                config_dict[split_line[0]] = int(split_line[1])

    return config_dict



import h5py
# import pathlib
import CyRSoXS as cy
import warnings
from .checkH5 import check_NumMat
from .reader import read_material, read_config
from .writer import write_opts, write_hdf5
from .visualizer import morphology_visualizer

import numpy as np
import xarray as xr
import sys
import os
import copy

from typing import Callable, TypeVar, ParamSpec

P = ParamSpec("P")
T = TypeVar("T")

def wraps(wrapper: Callable[P, T]) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    A decorator to preserve the original function's docstring.

    Args:
        wrapper: The wrapper function.

    Returns:
        A decorator that preserves the original function's docstring.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        """
        A decorator to preserve the original function's docstring.

        Args:
            func: The original function.

        Returns:
            The original function with its original docstring.
        """
        func.__doc__ = wrapper.__doc__
        return func

    return decorator

class Morphology:
    '''
    Object used to hold all the components necessary for a complete CyRSoXS morphology

    Attributes
    ----------
    input_mapping : dict
        A dictionary to handle specific configuration input types

    numMaterial : int
        Number of materials present in the morphology

    materials : dict
        A dictionary of Material objects

    PhysSize : float
        The physical size of each cubic voxel's three dimensions

    NumZYX : tuple or list
        Number of voxels in the Z, Y, and X directions (NumZ, NumY, NumX)

    config : dict
        A dictionary of configuration parameters for CyRSoXS

    create_cy_object : bool
        Boolean value that decides if the CyRSoXS objects are created upon instantiation

    simulated : bool
        Boolean value that tracks whether or not the simulation has been run

    Methods
    -------
    load_morph_hdf5(hdf5_file, create_cy_object=True)
        Class method that creates a Morphology object from a morphology HDF5 file

    create_inputData()
        Creates a CyRSoXS InputData object and populates it with parameters from self.config

    create_optical_constants()
        Creates a CyRSoXS RefractiveIndex object and populates it with optical constants from the materials dict

    create_voxel_data()
        Creates a CyRSoXS VoxelData object and populates it with the voxel information from the materials dict

    run(stdout=True,stderr=True, return_xarray=True, print_vec_info=False)
        Creates a CyRSoXS ScatteringPattern object if not already created, and submits all CyRSoXS objects to 
        run the simulation

    scattering_to_xarray(return_xarray=True,print_vec_info=False)
        Copies the CyRSoXS ScatteringPattern arrays to an xarray in the format used by PyHyperScattering for 
        further analysis
    '''

    # dict to deal with specific CyRSoXS input objects. dict structure inspired from David Ackerman's cyrsoxs-framework
    input_mapping = {'CaseType': ['setCaseType',[cy.CaseType.Default, cy.CaseType.BeamDivergence, cy.CaseType.GrazingIncidence]],
                     'MorphologyType': ['setMorphologyType', [cy.MorphologyType.EulerAngles, cy.MorphologyType.VectorMorphology]],
                     'EwaldsInterpolation': ['interpolationType', [cy.InterpolationType.NearestNeighour, cy.InterpolationType.Linear]],
                     'WindowingType': ['windowingType', [cy.FFTWindowing.NoPadding, cy.FFTWindowing.Hanning]],
                     'RotMask': ['rotMask', [False, True]],
                     'AlgorithmType': ['setAlgorithm', [0, 1]],
                     'ReferenceFrame': ['referenceFrame', [0, 1]]}

    config_default = {'CaseType': 0, 'Energies': [270.0], 'EAngleRotation': [0.0, 1.0, 0.0],
                      'MorphologyType': 0, 'AlgorithmType': 0, 'WindowingType': 0,
                      'RotMask': 0,
                      'ReferenceFrame': 1,
                      'EwaldsInterpolation': 1}

    def __init__(self, numMaterial, materials=None, PhysSize=None,
                 config={'CaseType': 0, 'MorphologyType': 0, 'Energies': [270.0], 'EAngleRotation': [0.0, 1.0, 0.0]}, create_cy_object=True):

        self._numMaterial = numMaterial
        self._PhysSize = PhysSize
        self.NumZYX = None
        self.inputData = None
        self.OpticalConstants = None
        self.voxelData = None
        self.scatteringPattern = None
        # add config keys and values to class dict
        for key in self.config_default:
            if key in config:
                self.__dict__['_'+key] = config[key]
            else:
                self.__dict__['_'+key] = self.config_default[key]

        # add materials
        self.materials = {}
        for i in range(1, self._numMaterial+1):
            if materials is None:
                self.materials[i] = Material(materialID=i)
            else:
                try:
                    self.materials[i] = materials[i].copy()
                    if i == 1:
                        self._Energies = materials[i].energies
                        self.NumZYX = materials[i].NumZYX
                except KeyError:
                    warnings.warn('numMaterial is greater than number of Material objects passed in. Creating empty Material')
                    self.materials[i] = Material(materialID=i)

        # flag denoting if Morphology has been simulated
        self._simulated = False

        if create_cy_object:
            self.create_update_cy()

    def __repr__(self):
        return f'Morphology (NumMaterial : {self.numMaterial}, PhysSize : {self.PhysSize})'

    @property
    def CaseType(self):
        return self._CaseType

    @CaseType.setter
    def CaseType(self, casevalue):
        if (casevalue != 0) & (casevalue != 1) & (casevalue !=2):
            raise ValueError('CaseType must be 0, 1, or 2')
        else:
            self._CaseType = casevalue

            if self.inputData:
                self.inputData.setCaseType(self.input_mapping['CaseType'][1][casevalue])

    @property
    def Energies(self):
        return self._Energies

    @Energies.setter
    def Energies(self, Elist):
        self._Energies = Elist

        if self.inputData:
            self.inputData.setEnergies(Elist)

    @property
    def EAngleRotation(self):
        return self._EAngleRotation

    @EAngleRotation.setter
    def EAngleRotation(self, anglelist):
        self._EAngleRotation = anglelist

        if self.inputData:
            self.inputData.setERotationAngle(StartAngle=anglelist[0],
                                             EndAngle=anglelist[2],
                                             IncrementAngle=anglelist[1])

    @property
    def MorphologyType(self):
        return self._MorphologyType

    @MorphologyType.setter
    def MorphologyType(self, value):
        if value != 0:
            raise ValueError('Only Euler Morphology is currently supported')
        else:
            self._MorphologyType = value

            if self.inputData:
                self.inputData.setMorphologyType(self.input_mapping['MorphologyType'][1][value])

    @property
    def AlgorithmType(self):
        return self._AlgorithmType

    @AlgorithmType.setter
    def AlgorithmType(self, value):
        if (value != 0) & (value != 1):
            raise ValueError('AlgorithmType must be 0 (communication minimizing) or 1 (memory minimizing).')
        else:
            self._AlgorithmType = value
            if self.inputData:
                self.inputData.setAlgorithm(AlgorithmID=value,MaxStreams=1)

    @property
    def WindowingType(self):
        return self._WindowingType

    @WindowingType.setter
    def WindowingType(self, value):
        if (value != 0) & (value != 1):
            raise ValueError('WindowingType must be 0 (None) or 1 (Hanning).')
        else:
            self._WindowingType = value

            if self.inputData:
                self.inputData.windowingType = self.input_mapping['WindowingType'][1][value]

    @property
    def RotMask(self):
        return self._RotMask

    @RotMask.setter
    def RotMask(self, value):
        if (value != 0) & (value != 1):
            raise ValueError('RotMask must be 0 (False) or 1 (True).')
        else:
            self._RotMask = value

            if self.inputData:
                self.inputData.rotMask = self._RotMask

    @property
    def EwaldsInterpolation(self):
        return self._EwaldsInterpolation

    @EwaldsInterpolation.setter
    def EwaldsInterpolation(self, value):
        if (value != 0) & (value != 1):
            raise ValueError('EwaldsInterpolation must be 0 (Nearest Neighbor) or 1 (Trilinear).')
        else:
            self._EwaldsInterpolation = value

            if self.inputData:
                self.inputData.interpolationType = self.input_mapping['EwaldsInterpolation'][1][value]

    @property
    def ReferenceFrame(self):
        return self._ReferenceFrame

    @ReferenceFrame.setter
    def ReferenceFrame(self, value):
        if (value != 0) & (value != 1):
            raise ValueError('ReferenceFrame must be 0 (Material Frame) or 1 (Lab Frame - Default).')
        else:
            self._ReferenceFrame = value

            if self.inputData:
                self.inputData.referenceFrame = self.input_mapping['ReferenceFrame'][1][value]

    @property
    def simulated(self):
        return self._simulated

    @property
    def PhysSize(self):
        return self._PhysSize

    @PhysSize.setter
    def PhysSize(self, val):
        if val < 0:
            raise ValueError('PhysSize must be greater than 0')
        self._PhysSize = float(val)
        # update inputData object
        if self.inputData:
            self.inputData.setPhysSize(self._PhysSize)

    @property
    def numMaterial(self):
        return self._numMaterial

    # @numMaterial.setter
    # def numMaterial(self, val):
    #     if val < 0:
    #         raise ValueError('numMaterial must be greater than 0')
    #     self._numMaterial = int(val)
    #     # if we change the number of materials and we have an inputData object, we need to recreate it with the new number of materials
    #     if self.inputData:
    #         self.create_inputData()
    #     if self.OpticalConstants:
    #         self.update_optical_constants()

    @property
    def config(self):
        return {key: self.__dict__['_'+key] for key in self.config_default}

    @config.setter
    def config(self, dict1):
        for key in dict1:
            if key in self.config_default:
                self.__dict__['_'+key] = dict1[key]
            else:
                warnings.warn(f'Key {key} not supported')

    @classmethod
    def load_morph_hdf5(cls, hdf5_file, create_cy_object=False):
        with h5py.File(hdf5_file, 'r') as f:
            if 'Euler_Angles' not in f.keys():
                raise KeyError('Only the Euler Angle convention is currently supported')
            # get number of materials in HDF5
            numMat = check_NumMat(f, morphology_type=0)
            PhysSize = f['Morphology_Parameters/PhysSize'][()]
            materials = dict()

            for i in range(numMat):
                materialID = i + 1
                # Load arrays directly without type conversion
                materials[materialID] = Material(
                    materialID=materialID,
                    Vfrac=f[f'Euler_Angles/Mat_{i+1}_Vfrac'][()],
                    S=f[f'Euler_Angles/Mat_{i+1}_S'][()],
                    theta=f[f'Euler_Angles/Mat_{i+1}_Theta'][()],
                    psi=f[f'Euler_Angles/Mat_{i+1}_Psi'][()],
                    NumZYX=f[f'Euler_Angles/Mat_{i+1}_Vfrac'][()].shape)

        return cls(numMat, materials=materials, PhysSize=PhysSize, create_cy_object=create_cy_object)

    def load_config(self, config_file):
        self.config = read_config(config_file)

    def load_matfile(self, matfile):
        return read_material(matfile)

    def create_inputData(self):
        self.inputData = cy.InputData(NumMaterial=self._numMaterial)
        # parse config dictionary and assign to appropriate places in inputData object
        self.config_to_inputData()

        if self.NumZYX is None:
            self.NumZYX = self.materials[1].NumZYX

        # only support ZYX ordering at the moment
        self.inputData.setDimensions(self.NumZYX, cy.MorphologyOrder.ZYX)

        if self.PhysSize is not None:
            self.inputData.setPhysSize(self.PhysSize)

        if not self.inputData.validate():
            warnings.warn('Validation failed. Double check inputData values')

    def create_optical_constants(self):
        self.OpticalConstants = cy.RefractiveIndex(self.inputData)
        self.update_optical_constants()        
        if not self.OpticalConstants.validate():
            warnings.warn('Validation failed. Double check optical constant values')

    def update_optical_constants(self):
        # Pre-allocate list to avoid repeated allocations
        all_constants = [None] * self._numMaterial
        for energy in self.Energies:
            # Update list in-place
            for ID in range(1, self.numMaterial+1):
                all_constants[ID-1] = self.materials[ID].opt_constants[energy]
            self.OpticalConstants.addData(OpticalConstants=all_constants, Energy=energy)

    def create_voxel_data(self):
        self.voxelData = cy.VoxelData(InputData=self.inputData)
        self.update_voxel_data()
        if not self.voxelData.validate():
            warnings.warn('Validation failed. Double check voxel data values')

    def update_voxel_data(self):
        for ID in range(1, self.numMaterial+1):
            mat = self.materials[ID]
            # Check types and only convert if needed
            s = mat.S if mat.S.dtype == np.float32 else mat.S.astype(np.float32)
            theta = mat.theta if mat.theta.dtype == np.float32 else mat.theta.astype(np.float32)
            psi = mat.psi if mat.psi.dtype == np.float32 else mat.psi.astype(np.float32)
            vfrac = mat.Vfrac if mat.Vfrac.dtype == np.float32 else mat.Vfrac.astype(np.float32)
            
            self.voxelData.addVoxelData(
                S=s,
                Theta=theta,
                Psi=psi,
                Vfrac=vfrac,
                MaterialID=ID)

    def config_to_inputData(self):
        for key in self.config:
            if key == "Energies":
                self.inputData.setEnergies(self.config[key])
            elif key == 'EAngleRotation':
                angles = self.config[key]
                self.inputData.setERotationAngle(StartAngle=float(angles[0]),
                                                 EndAngle=float(angles[2]),
                                                 IncrementAngle=float(angles[1]))
            elif key == 'AlgorithmType':
                self.inputData.setAlgorithm(AlgorithmID=self.config[key], MaxStreams=1)
            # if the key corresponds to one of the idiosyncratic methods, use this
            elif key in self.input_mapping.keys():
                func = getattr(self.inputData,self.input_mapping[key][0])
                if callable(func):
                    func(self.input_mapping[key][1][self.config[key]])
                # if the attribute is not callable, use input_mapping to set the attribute
                else:
                    setattr(self.inputData,
                            self.input_mapping[key][0],
                            self.input_mapping[key][1][self.config[key]])
            else:
                warnings.warn(f'{key} is currently not implemented')

    def create_update_cy(self):
        # create or update all CyRSoXS objects
        if self.inputData:
            self.config_to_inputData()
        else:
            self.create_inputData()

        # create or update OpticalConstants
        if self.OpticalConstants:
            self.update_optical_constants()            
        else:
            self.create_optical_constants()

        # create or udpate voxelData
        if self.voxelData:
            self.voxelData.reset()
            self.update_voxel_data()
        else:
            self.create_voxel_data()

    def write_to_file(self, fname, author='NIST'):
        _ = write_hdf5([[self.materials[i].Vfrac,
                        self.materials[i].S,
                        self.materials[i].theta,
                        self.materials[i].psi] for i in self.materials],
                        self.PhysSize, fname, self.MorphologyType, ordering='ZYX', author=author)

    # TODO : function to write a config.txt file from config dict
    def write_config(self,):
        pass

    def write_constants(self, path=None):
        for i in range(1, self._numMaterial+1):
            write_opts(self.materials[i].opt_constants, i, path)

    # submit to CyRSoXS
    def run(self, stdout=True, stderr=True, return_xarray=True, print_vec_info=False, validate=False):
        if validate:
            self.create_update_cy()

        # if we haven't created a ScatteringPattern object, create it now
        if not self.scatteringPattern:
            self.scatteringPattern = cy.ScatteringPattern(self.inputData)
        with cy.ostream_redirect(stdout=stdout, stderr=stderr):
            cy.launch(VoxelData=self.voxelData,
                      RefractiveIndexData=self.OpticalConstants,
                      InputData=self.inputData,
                      ScatteringPattern=self.scatteringPattern)
        self._simulated = True
        if return_xarray:
            return self.scattering_to_xarray(return_xarray=return_xarray, print_vec_info=print_vec_info)

    def scattering_to_xarray(self, return_xarray=True, print_vec_info=False):
        if self.simulated:
            if not print_vec_info:
                old_stdout = sys.stdout
                f = open(os.devnull, 'w')
                sys.stdout = f

            try:
                # Get data in [energy, NumY, NumX] shape
                scattering_data = self.scatteringPattern.writeAllToNumpy(kID=0)
                
                # Pre-compute FFT frequencies once
                ny, nx = self.NumZYX[1:]
                d = self.PhysSize
                
                # Calculate q-vectors
                qy = 2*np.pi*np.fft.fftshift(np.fft.fftfreq(ny, d=d))
                qx = 2*np.pi*np.fft.fftshift(np.fft.fftfreq(nx, d=d))
                
                # Create DataArray with pre-computed coordinates
                scattering_data = xr.DataArray(
                    scattering_data,
                    dims=['energy', 'qy', 'qx'],
                    coords={'qy': qy, 'qx': qx, 'energy': self.config['Energies']}
                )
                
                if return_xarray:
                    return scattering_data
                else:
                    self.scattering_data = scattering_data
                    
            finally:
                if not print_vec_info:
                    sys.stdout = old_stdout
                    f.close()
        else:
            warnings.warn('You haven\'t run your simulation yet')

    # TODO : restructure to have a single checkH5 engine for both NRSS and
    # command line formats
    def check_materials(self, quiet=True):
        # Vectorized sum of volume fractions
        Vfrac_sum = sum(mat.Vfrac for mat in self.materials.values())
        assert np.allclose(Vfrac_sum, 1), 'Total material volume fractions do not sum to 1'
        del Vfrac_sum  # Free memory immediately
        
        # Check each material's properties
        for i, mat in self.materials.items():
            # Validate value ranges and types in one pass
            for name, arr in {
                'S': mat.S,
                'Vfrac': mat.Vfrac,
                'theta': mat.theta,
                'psi': mat.psi
            }.items():
                # Check for NaNs and float type in one operation
                if not np.issubdtype(arr.dtype, np.floating):
                    raise AssertionError(f'Material {i} {name} is not of type float')
                
                if np.any(np.isnan(arr)):
                    raise AssertionError(f'NaNs are present in Material {i} {name}')
                
                # Check bounds for S and Vfrac
                if name in ('S', 'Vfrac'):
                    if not np.all((arr >= 0) & (arr <= 1)):
                        raise AssertionError(f'Material {i} {name} value(s) does not lie between 0 and 1')
        
        if not quiet:
            print('All material checks have passed')

    @wraps(morphology_visualizer)
    def visualize_materials(self, *args,**kwargs):
        return morphology_visualizer(self, *args,**kwargs)
    visualize_materials.__doc__ = morphology_visualizer.__doc__


    def validate_all(self, quiet=True):
        self.check_materials(quiet=quiet)
        input_check = self.inputData.validate()
        opt_const_check = self.OpticalConstants.validate()
        voxel_check = self.voxelData.validate()
        assert (input_check), 'CyRSoXS object inputData validation has failed'
        assert (opt_const_check), 'CyRSoXS object OpticalConstants validation has failed'
        assert (voxel_check), 'CyRSoXS object voxelData validation has failed'
        if not quiet:
            print('All objects have been validated successfully. You can run your simulation')


class OpticalConstants:
    '''
    Object to hold dielectric optical constants in a format compatible with CyRSoXS

    Attributes
    ----------

    energies : list or array
        List of energies
    opt_constants : dict
        Dictionary of optical constants, where each energy is a key in the dict
    name : str
        String identifying the element or material for these optical constants

    Methods
    -------
    calc_constants(energies, reference_data, name='unkown')
        Interpolates optical constant data to the list of energies provided
    load_matfile(matfile, name='unknown')
        Creates an OpticalConstants object from a previously written MaterialX.txt file
    create_vacuum(energies)
        Convenience function to populate zeros for all optical constants

    '''

    def __init__(self, energies, opt_constants=None, name='unknown'):
        self.energies = energies
        self.opt_constants = opt_constants
        self.name = name
        if self.name == 'vacuum':
            self.create_vacuum(energies)

    def __repr__(self):
        return f'OpticalConstants (Material : {self.name}, Number of Energies : {len(self.energies)})'

    @classmethod
    def calc_constants(cls, energies, reference_data, name='unknown'):
        deltabeta = dict()
        for energy in energies:
            dPara = np.interp(energy, reference_data['Energy'], reference_data['DeltaPara'])
            bPara = np.interp(energy, reference_data['Energy'], reference_data['BetaPara'])
            dPerp = np.interp(energy, reference_data['Energy'], reference_data['DeltaPerp'])
            bPerp = np.interp(energy, reference_data['Energy'], reference_data['BetaPerp'])
            deltabeta[energy] = [dPara, bPara, dPerp, bPerp]
        return cls(energies, deltabeta, name=name)

    @classmethod
    def load_matfile(cls, matfile, name='unknown'):
        energies, deltabeta = read_material(matfile)
        return cls(energies, deltabeta, name=name)

    def create_vacuum(self, energies):
        deltabeta = dict()
        for energy in energies:
            deltabeta[energy] = [0.0, 0.0, 0.0, 0.0]
        self.energies = energies
        self.opt_constants = deltabeta


class Material(OpticalConstants):
    '''
    Object to hold the voxel-level data for a CyRSoXS morphology. Inherits from the OpticalConstants class.

    Attributes
    ----------
    materialID : int
        Integer value denoting the material number. Used in CyRSoXS
    Vfrac : ndarray
        Volume fractions for a Material
    theta : ndarray
        The second Euler angle (ZYZ convention)
    psi : ndarray
        The third Euler angle (ZYZ convention)
    NumZYX : tuple or list
        Dimensions of the Material arrays (NumZ, NumY, NumX)
    name : str
        Name of the Material (e.g. 'Polystyrene')

    '''

    def __init__(self, materialID=1, Vfrac=None, S=None, theta=None, psi=None,
                 NumZYX=None, energies=None, opt_constants=None, name=None):
        self.materialID = materialID
        # Store arrays as-is, type conversion will happen when needed
        self.Vfrac = Vfrac
        self.S = S
        self.theta = theta
        self.psi = psi
        self.NumZYX = NumZYX
        self.name = name
        self.energies = energies
        self.opt_constants = opt_constants
        if self.NumZYX is None:
            try:
                self.NumZYX = Vfrac.shape
            except AttributeError:
                pass

        if (energies is None) & (opt_constants is not None):
            self.energies = list(opt_constants.keys())

        super().__init__(self.energies, self.opt_constants, name=name)

    def __repr__(self):
        return f'Material (Name : {self.name}, ID : {self.materialID}, Shape : {self.NumZYX})'

    def __copy__(self):
        return Material(materialID=self.materialID,
                        Vfrac=self.Vfrac.copy(),
                        S=self.S.copy(),
                        theta=self.theta.copy(),
                        psi=self.psi.copy(),
                        NumZYX=self.NumZYX,
                        energies=self.energies,
                        opt_constants=self.opt_constants,
                        name=self.name)

    def copy(self):
        return copy.copy(self)



import numpy as np
import pyFAI,pyFAI.azimuthalIntegrator
import warnings
import xarray as xr
import h5py
from skimage.transform import warp_polar

# global constants for this file
h = 4.135667696e-15 #Planck's constant [eV*s]
c = 299792458 # speed of light [m/s]

def remesh(img):
    remeshed = warp_polar(img,center=(img.shape[0]//2, img.shape[1]//2))
    q = np.sqrt(img.Qy**2+img.Qx**2)
    output_q = np.linspace(0,np.amax(q),remeshed.shape[1])
    output_chi = np.linspace(0.5,359.5,360)
    da_remeshed = xr.DataArray(remeshed,dims=['chi','q'],coords=dict(q=output_q,chi=output_chi))
    return da_remeshed

def read_img_h5(fname,PhysSize=5):
    with h5py.File(fname,'r') as h5:
        img = h5['K0']['projection'][()]
    Qx = 2.0*np.pi*np.fft.fftshift(np.fft.fftfreq(img.shape[1],d=PhysSize))
    Qy = 2.0*np.pi*np.fft.fftshift(np.fft.fftfreq(img.shape[0],d=PhysSize))
    da_img = xr.DataArray(img,dims=['Qy','Qx'],coords=dict(Qx=Qx,Qy=Qy))
    return da_img


def read_config(fname):
    config = {}
    with open(fname) as f:
        for line in f:
            key,value = line.split('=')
            key = key.strip()
            value = value.split(';')[0].strip()
            if key in ['NumX','NumY','NumZ','NumThreads','EwaldsInterpolation','WindowingType']:
                value = int(value)
            elif key in ['RotMask','WriteVTI']:
                value = bool(value)
            else:
                value = float(value)
            config[key] = value
    return config




from copy import copy
import numpy as np
import cupy as cp
from NRSS.morphology import Material, Morphology, wraps
from PyHyperScattering.integrate import WPIntegrator
import pickle, lzma
from pathlib import Path
import glob
from tqdm import tqdm
from pprint import pprint
from matplotlib import colormaps, gridspec, rc, ticker
from matplotlib import pyplot as plt
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize
from xarray import DataArray
from dataclasses import dataclass, field


@dataclass
class Arguments():
    """
    Dataclass for storing parameters for the core-shell spheres decay simulation
    """
    r: int = 512
    """
    Number of voxels in YX dimension of the simulation box
    """
    d: int = 32
    """
    Number of voxels in Z dimension of the simulation box
    """
    vx_size: float = 2.5
    """
    Size of each voxel in the simulation box in nm
    """
    ra: float = 9.9
    """
    Radius of the core particle
    """
    t: float = 7.3
    """
    Thickness of the oriented corona
    """
    S_0: float = 0.11
    """
    orientation immediately near the core particle
    """
    S_range_max: float = 0.11
    """
    orientation magnitude maximum sed for S visualizations
    """
    S_cmap: str = "afmhot"
    """
    orientation immediately near the core particle
    """
    decay_order: float = 0.421
    """
    Decay order for the core-shell spheres
    """
    energies: np.ndarray = field(default_factory=lambda: np.array([260, 270, 285]))
    """
    Energies at which to simulate the scattering
    """
    coord_list: list = field(default_factory=list)
    """
    List of coordinates for the core-shell spheres
    """
    return_2d: bool = False
    """
    whether to return 2D pattern simulations in result object
    """

@dataclass
class Visualization():
    """
    Dataclass for storing visualization parameters and results
    """
    outputmat: list = field(default_factory=list)
    """
    List of materials to output in the visualization
    """
    outputplot: list = field(default_factory=list)
    """
    List of plots to output in the visualization
    """
    img_dict: dict = field(default_factory=dict)
    """
    List of images output from the visualization
    """
    
    @wraps(Morphology.visualize_materials)
    def get_visualizations(self, morph, **kwargs):
        """
        Get visualization results for the given morphology
        
        Parameters
        ----------
        morph : Morphology
            Morphology object to visualize
        **kwargs
            Additional keyword arguments to pass to `morph.visualize_materials`
        
        Returns
        -------
        self : Visualization
            The same Visualization object with the results stored as attributes
        """
        kwargs["runquiet"] = True
        kwargs["outputmat"] = self.outputmat
        kwargs["outputplot"] = self.outputplot

        visualization_results = morph.visualize_materials(
        **kwargs
        )
        self.img_dict = {}
        for i, mat in enumerate(self.outputmat):
            for j, vizplot in enumerate(self.outputplot):
                self.img_dict[f"mat_{mat}_{vizplot}"] = visualization_results[i*len(self.outputplot) + j]
        return self

@dataclass
class Results():
    twod: DataArray = field(default_factory=DataArray) # 2D image data from the simulation
    I: DataArray = field(default_factory=DataArray) # Intensity data from the simulation
    A: DataArray  = field(default_factory=DataArray) # Anisotropy data from the simulation
    viz: list = field(default_factory=list)  # Visualization data from the simulation


def coreshell_spheres_decay_euler_old(args):

    r = args.r
    d = args.d
    vx_size = args.vx_size
    ra_vx = args.ra / args.vx_size
    t_vx = args.t / args.vx_size
    phi_iso = 1 - args.S_0
    decay_order = args.decay_order
    l = args.coord_list

    # Initialize boolean masks for material presence
    a_b = cp.full((d, r, r), False)  # core
    b_b = cp.full((d, r, r), False)  # shell
    c_b = cp.full((d, r, r), False)  # everything else

    # Initialize orientation components of shell
    b_x = cp.zeros([d, r, r])  # shell
    b_y = cp.zeros([d, r, r])  # shell
    b_z = cp.zeros([d, r, r])  # shell

    # Initialize volume fractions of each material - euler
    vf_a = cp.zeros([d, r, r])  # core
    vf_b = cp.zeros([d, r, r])  # shell
    vf_c = cp.zeros([d, r, r])  # everything else

    # Initialize euler angles and aligned fraction - euler morphology - ZXZ rotations
    SE_b = cp.zeros([d, r, r])
    theta_b = cp.zeros([d, r, r])  # theta of shell
    psi_b = cp.zeros([d, r, r])  # psi of shell

    # Initialize intermediate orientation components of shell
    b_xi = cp.zeros([d, r, r])  # shell
    b_yi = cp.zeros([d, r, r])  # shell
    b_zi = cp.zeros([d, r, r])  # shell

    z, y, x = cp.ogrid[0:d:1, 0:r:1, 0:r:1]

    for p in l:
        mf = (x - p[0]) ** 2 + (y - p[1]) ** 2 + (z - 15) ** 2 
        mask = mf <= ra_vx ** 2
        a_b = cp.logical_or(a_b, mask)
        mask = (a_b == False) & (mf <= (ra_vx + t_vx) ** 2)
        b_b = cp.logical_or(b_b, mask)

        b_xi = ((x - p[0]) + y * 0 + z * 0) * mask
        b_yi = (x * 0 + (y - p[1]) + z * 0) * mask
        b_zi = (x * 0 + y * 0 + (z - 15)) * mask

        b_x = b_x + b_xi * (b_x == 0)
        b_y = b_y + b_yi * (b_y == 0)
        b_z = b_z + b_zi * (b_z == 0)

    b_b = (a_b == False) & (b_b == True)
    b_x = b_x * b_b
    b_y = b_y * b_b
    b_z = b_z * b_b

    c_b = (a_b == False) & (b_b == False)

    b_t = (b_x ** 2 + b_y ** 2 + b_z ** 2) ** 0.5

    b_1 = (ra_vx) * b_b

    vf_b = 1 * b_b
    SE_b = b_b * (1 - phi_iso) * (b_1 / b_t) ** (decay_order)
    vf_a = 1 * a_b
    vf_c = 1 * c_b

    SE_b = cp.nan_to_num(SE_b)

    theta_b = cp.arctan2(((b_x) ** 2 + (b_y) ** 2) ** 0.5, b_z)
    psi_b = cp.arctan2(b_y, b_x)

    vf_a = cp.asnumpy(vf_a)
    vf_b = cp.asnumpy(vf_b)
    vf_c = cp.asnumpy(vf_c)
    SE_b = cp.asnumpy(SE_b)
    theta_b = cp.asnumpy(theta_b)
    psi_b = cp.asnumpy(psi_b)

    a_zeros = np.zeros(vf_a.shape, dtype=np.float32)

    mat1_poly = Material(
        materialID=1,
        Vfrac=(vf_b).astype(np.float32),
        S=SE_b.astype(np.float32),
        theta=theta_b.astype(np.float32),
        psi=psi_b.astype(np.float32),
        NumZYX=vf_a.shape,
        energies=args.energies,
        opt_constants=args.oc_polymer.opt_constants,
        name="PS",
    )

    mat2_particle = Material(
        materialID=2,
        Vfrac=vf_a.astype(np.float32),
        S=a_zeros,
        theta=a_zeros,
        psi=a_zeros,
        NumZYX=vf_a.shape,
        energies=args.energies,
        opt_constants=args.oc_particle.opt_constants,
        name="particle",
    )

    # u = np.random.uniform(size = vf_c.shape)
    # v = np.random.uniform(size =vf_c.shape)
    # theta_rand = 2*np.pi*u
    # psi_rand = np.arccos(2*v-1)

    mat3_poly = Material(
        materialID=3,
        Vfrac=(vf_c).astype(np.float32),
        S=a_zeros.astype(np.float32),
        theta=a_zeros.astype(np.float32),
        psi=a_zeros.astype(np.float32),
        NumZYX=a_zeros.shape,
        energies=args.energies,
        opt_constants=args.oc_polymer.opt_constants,
        name="PS",
    )

    morph = Morphology(
        3,
        {1: mat1_poly, 2: mat2_particle, 3: mat3_poly},
        PhysSize=args.vx_size,
    )

    # Clean up
    del(a_b, b_b, c_b, b_x, b_y, b_z, b_xi, b_yi, b_zi, b_t, b_1, SE_b, theta_b, psi_b, vf_a, vf_b, vf_c, a_zeros, mat1_poly, mat2_particle)

    return morph


def coreshell_spheres_decay_euler(args):
    """
    Creates a morphology with a core-shell particle and a polymer matrix.

    Parameters
    ----------
    args : Arguments
        A dataclass containing the parameters for the morphology

    Returns
    -------
    morph : Morphology
        The created morphology
    """
    # Initialize boolean masks for material presence
    core_mask = cp.zeros((args.d, args.r, args.r), dtype=bool)
    shell_mask = cp.zeros((args.d, args.r, args.r), dtype=bool)
    polymer_mask = cp.ones((args.d, args.r, args.r), dtype=bool)

    # Initialize orientation components of shell
    shell_x = cp.zeros((args.d, args.r, args.r))
    shell_y = cp.zeros((args.d, args.r, args.r))
    shell_z = cp.zeros((args.d, args.r, args.r))

    # Initialize volume fractions of each material
    core_volume_fraction = cp.zeros((args.d, args.r, args.r))
    shell_volume_fraction = cp.zeros((args.d, args.r, args.r))
    polymer_volume_fraction = cp.zeros((args.d, args.r, args.r))

    # Initialize euler angles and aligned fraction - euler morphology - ZXZ rotations
    shell_S = cp.zeros((args.d, args.r, args.r))
    shell_theta = cp.zeros((args.d, args.r, args.r))
    shell_psi = cp.zeros((args.d, args.r, args.r))

    # Initialize intermediate orientation components of shell
    shell_xi = cp.zeros((args.d, args.r, args.r))
    shell_yi = cp.zeros((args.d, args.r, args.r))
    shell_zi = cp.zeros((args.d, args.r, args.r))

    z, y, x = cp.ogrid[0 : args.d : 1, 0 : args.r : 1, 0 : args.r : 1]

    for point in args.coord_list:
        squared_distance_from_point = (
            (x - point[0]) ** 2 + (y - point[1]) ** 2 + (z - 15) ** 2
        )  # calculate the squared distance from each point in the grid to the current particle
        mask = squared_distance_from_point <= (args.ra / args.vx_size) ** 2  # create a mask for the particle core
        core_mask = cp.logical_or(core_mask, mask)  # update the core mask
        mask = (core_mask == False) & (
            squared_distance_from_point
            <= (args.ra / args.vx_size + args.t / args.vx_size) ** 2
        )  # create a mask for the particle shell
        shell_mask = cp.logical_or(shell_mask, mask)  # update the shell mask

        shell_xi = ((x - point[0]) + y * 0 + z * 0) * mask  # calculate the x-component of the shell orientation
        shell_yi = (x * 0 + (y - point[1]) + z * 0) * mask  # calculate the y-component of the shell orientation
        shell_zi = (x * 0 + y * 0 + (z - 15)) * mask  # calculate the z-component of the shell orientation

        shell_x += shell_xi * (shell_x == 0)  # update the x-component of the shell orientation
        shell_y += shell_yi * (shell_y == 0)  # update the y-component of the shell orientation
        shell_z += shell_zi * (shell_z == 0)  # update the z-component of the shell orientation

    shell_mask = (core_mask == False) & (shell_mask == True)
    shell_x *= shell_mask
    shell_y *= shell_mask
    shell_z *= shell_mask

    polymer_mask = (core_mask == False) & (shell_mask == False)

    shell_t = (shell_x**2 + shell_y**2 + shell_z**2) ** 0.5  # calculate the magnitude of the shell orientation

    shell_1 = (args.ra / args.vx_size) * shell_mask  # calculate the radial distance from the particle center to the shell

    shell_volume_fraction = 1 * shell_mask  # calculate the volume fraction of the shell
    shell_S = shell_mask * args.S_0 * (shell_1 / shell_t) ** (args.decay_order)  # calculate the aligned fraction of the shell
    core_volume_fraction = 1 * core_mask  # calculate the volume fraction of the particle core
    polymer_volume_fraction = 1 * polymer_mask  # calculate the volume fraction of the polymer matrix

    shell_S = cp.nan_to_num(shell_S)

    shell_theta = cp.arctan2(((shell_x) ** 2 + (shell_y) ** 2) ** 0.5, shell_z)  # calculate the theta angle of the shell orientation
    shell_psi = cp.arctan2(shell_y, shell_x)  # calculate the psi angle of the shell orientation

    core_volume_fraction = cp.asnumpy(core_volume_fraction)
    shell_volume_fraction = cp.asnumpy(shell_volume_fraction)
    polymer_volume_fraction = cp.asnumpy(polymer_volume_fraction)
    shell_S = cp.asnumpy(shell_S)
    shell_theta = cp.asnumpy(shell_theta)
    shell_psi = cp.asnumpy(shell_psi)

    a_zeros = np.zeros(core_volume_fraction.shape, dtype=np.float32)

    mat1_poly = Material(
        materialID=1,
        Vfrac=(shell_volume_fraction).astype(np.float32),
        S=shell_S.astype(np.float32),
        theta=shell_theta.astype(np.float32),
        psi=shell_psi.astype(np.float32),
        NumZYX=core_volume_fraction.shape,
        energies=args.energies,
        opt_constants=args.oc_polymer.opt_constants,
        name="PS",
    )

    mat2_particle = Material(
        materialID=2,
        Vfrac=core_volume_fraction.astype(np.float32),
        S=a_zeros,
        theta=a_zeros,
        psi=a_zeros,
        NumZYX=core_volume_fraction.shape,
        energies=args.energies,
        opt_constants=args.oc_particle.opt_constants,
        name="particle",
    )

    mat3_poly = Material(
        materialID=3,
        Vfrac=(polymer_volume_fraction).astype(np.float32),
        S=a_zeros.astype(np.float32),
        theta=a_zeros.astype(np.float32),
        psi=a_zeros.astype(np.float32),
        NumZYX=a_zeros.shape,
        energies=args.energies,
        opt_constants=args.oc_polymer.opt_constants,
        name="PS",
    )

    morph = Morphology(
        3,
        {1: mat1_poly, 2: mat2_particle, 3: mat3_poly},
        PhysSize=args.vx_size,
    )

    del (
        core_mask,
        shell_mask,
        polymer_mask,
        shell_x,
        shell_y,
        shell_z,
        shell_xi,
        shell_yi,
        shell_zi,
        core_volume_fraction,
        shell_volume_fraction,
        polymer_volume_fraction,
        shell_S,
        shell_theta,
        shell_psi,
        a_zeros,
        mat1_poly,
        mat2_particle,
        mat3_poly,
    )
    # a cupy helper function to ensure that unused memory is freed
    cp._default_memory_pool.free_all_blocks()
    return morph

def coreshell_spheres_decay_euler_16bit(args):
    """
    Creates a morphology with a core-shell particle and a polymer matrix.

    Parameters
    ----------
    args : Arguments
        A dataclass containing the parameters for the morphology

    Returns
    -------
    morph : Morphology
        The created morphology
    """
    # Initialize boolean masks for material presence
    core_mask = cp.zeros((args.d, args.r, args.r), dtype=bool)
    shell_mask = cp.zeros((args.d, args.r, args.r), dtype=bool)
    polymer_mask = cp.ones((args.d, args.r, args.r), dtype=bool)

    # Initialize orientation components of shell
    shell_x = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)
    shell_y = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)
    shell_z = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)

    # Initialize volume fractions of each material
    core_volume_fraction = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)
    shell_volume_fraction = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)
    polymer_volume_fraction = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)

    # Initialize euler angles and aligned fraction - euler morphology - ZXZ rotations
    shell_S = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)
    shell_theta = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)
    shell_psi = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)

    # Initialize intermediate orientation components of shell
    shell_xi = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)
    shell_yi = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)
    shell_zi = cp.zeros((args.d, args.r, args.r), dtype=cp.float16)

    z,y,x = cp.ix_(cp.arange(args.d, dtype=cp.float16), cp.arange(args.r, dtype = cp.float16), cp.arange(args.r, dtype = cp.float16))

    for point in args.coord_list:
        squared_distance_from_point = (
            (x - cp.float16(point[0])) ** 2 + (y - cp.float16(point[1])) ** 2 + (z - 15) ** 2
        )  # calculate the squared distance from each point in the grid to the current particle
        mask = squared_distance_from_point <= (cp.float16(args.ra / args.vx_size)) ** 2  # create a mask for the particle core
        core_mask = cp.logical_or(core_mask, mask)  # update the core mask
        mask = (core_mask == False) & (
            squared_distance_from_point
            <= cp.float16(args.ra / args.vx_size + args.t / args.vx_size) ** 2
        )  # create a mask for the particle shell
        shell_mask = cp.logical_or(shell_mask, mask)  # update the shell mask

        shell_xi = ((x - cp.float16(point[0])) + y * 0 + z * 0) * mask  # calculate the x-component of the shell orientation
        shell_yi = (x * 0 + (y - cp.float16(point[1])) + z * 0) * mask  # calculate the y-component of the shell orientation
        shell_zi = (x * 0 + y * 0 + (z - 15)) * mask  # calculate the z-component of the shell orientation

        shell_x += shell_xi * (shell_x == 0)  # update the x-component of the shell orientation
        shell_y += shell_yi * (shell_y == 0)  # update the y-component of the shell orientation
        shell_z += shell_zi * (shell_z == 0)  # update the z-component of the shell orientation

    shell_mask = (core_mask == False) & (shell_mask == True)
    shell_x *= shell_mask
    shell_y *= shell_mask
    shell_z *= shell_mask

    polymer_mask = (core_mask == False) & (shell_mask == False)

    shell_t = (shell_x**2 + shell_y**2 + shell_z**2) ** 0.5  # calculate the magnitude of the shell orientation

    shell_1 = (args.ra / args.vx_size) * shell_mask  # calculate the radial distance from the particle center to the shell

    shell_volume_fraction = 1 * shell_mask  # calculate the volume fraction of the shell
    shell_S = shell_mask * cp.float16(args.S_0) * (shell_1 / shell_t) ** cp.float16(args.decay_order)  # calculate the aligned fraction of the shell
    core_volume_fraction = 1 * core_mask  # calculate the volume fraction of the particle core
    polymer_volume_fraction = 1 * polymer_mask  # calculate the volume fraction of the polymer matrix

    shell_S = cp.nan_to_num(shell_S)

    shell_theta = cp.arctan2(((shell_x) ** 2 + (shell_y) ** 2) ** 0.5, shell_z)  # calculate the theta angle of the shell orientation
    shell_psi = cp.arctan2(shell_y, shell_x)  # calculate the psi angle of the shell orientation

    core_volume_fraction = cp.asnumpy(core_volume_fraction)
    shell_volume_fraction = cp.asnumpy(shell_volume_fraction)
    polymer_volume_fraction = cp.asnumpy(polymer_volume_fraction)
    shell_S = cp.asnumpy(shell_S)
    shell_theta = cp.asnumpy(shell_theta)
    shell_psi = cp.asnumpy(shell_psi)

    a_zeros = np.zeros(core_volume_fraction.shape, dtype=np.float32)

    mat1_poly = Material(
        materialID=1,
        Vfrac=(shell_volume_fraction).astype(np.float32),
        S=shell_S.astype(np.float32),
        theta=shell_theta.astype(np.float32),
        psi=shell_psi.astype(np.float32),
        NumZYX=core_volume_fraction.shape,
        energies=args.energies,
        opt_constants=args.oc_polymer.opt_constants,
        name="PS",
    )

    mat2_particle = Material(
        materialID=2,
        Vfrac=core_volume_fraction.astype(np.float32),
        S=a_zeros,
        theta=a_zeros,
        psi=a_zeros,
        NumZYX=core_volume_fraction.shape,
        energies=args.energies,
        opt_constants=args.oc_particle.opt_constants,
        name="particle",
    )

    mat3_poly = Material(
        materialID=3,
        Vfrac=(polymer_volume_fraction).astype(np.float32),
        S=a_zeros.astype(np.float32),
        theta=a_zeros.astype(np.float32),
        psi=a_zeros.astype(np.float32),
        NumZYX=a_zeros.shape,
        energies=args.energies,
        opt_constants=args.oc_polymer.opt_constants,
        name="PS",
    )

    morph = Morphology(
        3,
        {1: mat1_poly, 2: mat2_particle, 3: mat3_poly},
        PhysSize=args.vx_size,
    )

    del (
        core_mask,
        shell_mask,
        polymer_mask,
        shell_x,
        shell_y,
        shell_z,
        shell_xi,
        shell_yi,
        shell_zi,
        core_volume_fraction,
        shell_volume_fraction,
        polymer_volume_fraction,
        shell_S,
        shell_theta,
        shell_psi,
        a_zeros,
        mat1_poly,
        mat2_particle,
        mat3_poly,
    )
    # a cupy helper function to ensure that unused memory is freed
    cp._default_memory_pool.free_all_blocks()
    return morph

def coreshell_spheres_run(args):

    # notice how we're passing args to the adsorbed_polymer_morphology function
    # we could also have args that affect how the model is run (but we don't right now)

    # we start right away with the adsorbed_polymer_morphology function
    # Create morphology object
    morph = coreshell_spheres_decay_euler(args)
    
    #this creates a single-particle model to aid visualization
    args_mdl = copy(args) # args.copy()
    args_mdl.coord_list = [[args.r/2, args.r/2]]
    
    morph_mdl = coreshell_spheres_decay_euler(args_mdl)    

    # FFT "window" parameter
    morph.inputData.windowingType = 0 #cy.FFTWindowing.Hanning

    morph.EAngleRotation = [0.0, 2.0, 360.0]

    # we will keep the validator in here. Always good to check the morphology your function made
    morph.validate_all(quiet=False)

    # we will use a dataclass for holding the results
    results = Results()

    # this actually runs the model
    data = morph.run(stdout=True, stderr=False)

    if args.return_2d:    
        results.twod = data.copy() # note that this gets deleted, so copy is appropriate

    # the PyHyperScattering part for interpreting the model result
    integrator = WPIntegrator(use_chunked_processing=False)
    remeshed_data = integrator.integrateImageStack(data)


        
    results.I = remeshed_data.mean(dim="chi")
    results.A = remeshed_data.rsoxs.AR(chi_width=5)

    # note, in addition to the model output, you could also return other things like the visualization results (which you might want to see, for parameter sweeps)
    # morphology.visualize_materials has a huge number of potential arguments to tailor what it shows about the model and how it shows it

    results.viz = Visualization(outputmat = [1], outputplot=['vfrac', 'S', 'psi', 'theta'])
    results.viz.get_visualizations(
        morph_mdl,
        subsample = 32,
        z_slice= 16,
        outputmat = [1],
        outputplot=['vfrac', 'S', 'psi', 'theta'],
        S_range=[[0, 0.2]],
        runquiet=True,
        outputaxes = False,
        S_cmap = args.S_cmap,
    )

    # also, it's a great idea to store the args dictionary in the results
    # that creates a durable record of what the parameters were when you simulated
    results.args = args

    # you could choose to return a variety of information.
    # for example, you could return the data object and/or the remeshed_data object
    # if you wanted to see full 2D patterns
    # I'm choosing only I and A, which are reductions of the data, but ones focused on our inquiry
    # I'm also returning the visualizations created above

    # note we return the results dictionary, which has I, A, and visualizations stored in it
    del(data, args_mdl, morph, morph_mdl, integrator, remeshed_data)
    cp._default_memory_pool.free_all_blocks()
    return results

def run_sweep(args, swept_arg, swept_range):
    """
    Runs the model for a range of values for a single argument and saves the results to pickle files.

    Parameters:
        args (dict): A dictionary containing the arguments for running the model.
        swept_arg (str): The name of the argument to sweep. Choose from "t", "S_0", and "decay_order" 
        swept_range (list): A list of values to sweep over.
    """
    # loop through the range of values
    for i, arg_value in enumerate(swept_range):
        # set the value of the argument to the current value in the range
        setattr(args, swept_arg, arg_value)
        # run the model with the current argument
        results = coreshell_spheres_run(args)
        # create a filename for the pickle file
        filename = f"{swept_arg}_{i:03d}.pkl"
        # create a path to the pickle directory
        pickle_path = Path.cwd() / "pickles"
        # create the full path and name of the pickle file
        savefile_path_and_name = pickle_path / filename
        # save the results to the pickle file
        with lzma.open(savefile_path_and_name, "wb") as handle:
            pickle.dump(results, handle, protocol=pickle.HIGHEST_PROTOCOL)

def import_pickles(picklePath, sstring):
    # below is a "docstring" and it is used to describe the function
    """
    Imports pickle files from a specified directory that match a given string.

    Parameters:
        picklePath (str): The path to the directory containing the pickle files.
        sstring (str): The string that the file names must contain.

    Returns:
        list: A list of objects representing the loaded pickle data.
    """
    # initialize an empty list to store the results
    results = []
    # use glob to find all the files in the directory that match the string
    # note the wildcard "*" in the glob function
    file_list = glob.glob(f"{picklePath}/{sstring}*")
    # show us those filenames
    pprint(file_list)
    # sort them - this will put them in the order they were created
    file_list.sort()
    # loop through the files. Notice how tqdm works - it actually is a function around the list that is iterated through
    for file_name in tqdm(file_list, colour = "blue"):
        # open the file using lzma
        with lzma.open(file_name, "rb") as pickle_file:
            # load the pickle file and append it to the results list
            results.append(pickle.load(pickle_file))
    # return the whole results list
    return results


def plot_sweep(swept_arg):
    """
    Plots the results of a sweep of the argument `swept_arg`.

    Parameters:
        swept_arg (str): The name of the argument to sweep. Choose from "t", "S_0", and "decay_order" The results should previously have been saved to pickle files using the `run_sweep` function.

    Returns:
        None
    """
    results = import_pickles(Path.cwd() / "pickles", swept_arg)
    num_viz = len(results)
    viz_col = np.ceil(num_viz / 3).astype(int)

    # font = {
    #     "family": "sans-serif",
    #     "sans-serif": "Arial",
    #     "weight": "regular",
    #     "size": 8,
    # }

    # rc("font", **font)

    # fig, _ = plt.subplots()
    fig, _ = plt.subplots(figsize=(7,8), dpi = 100)

    gs = gridspec.GridSpec(
        nrows=2,
        ncols=1,
        figure=fig,
        width_ratios=[1],
        height_ratios=[0.6, 0.4],
        wspace=0.1,
        hspace=0.15,
    )
    ax1 = plt.subplot(gs[0:1])

    gs_lower =  gridspec.GridSpecFromSubplotSpec(
        nrows=1,
        ncols=3,
        subplot_spec=gs[1:2],
        width_ratios=[0.9, 0.02, 0.08],
        wspace=0,
        hspace=0,
    )

    ax_cbar = plt.subplot(gs_lower[1:2])
    
    ax_backdrop = plt.subplot(gs_lower[0:1])
    ax_backdrop.set_facecolor("black")
    ax_backdrop.tick_params(left = False, right = False , labelleft = False , 
                labelbottom = False, bottom = False) 
    
    gs_viz =  gridspec.GridSpecFromSubplotSpec(
        nrows=3,
        ncols=viz_col,
        subplot_spec=gs_lower[0:1],
        wspace=0,
        hspace=0,
    )
    # axs = plt.subplots(1,18, gs[1:2])

    norm = Normalize(vmin=0, vmax=results[0].args.S_range_max)
    cbar = fig.colorbar(ScalarMappable(norm=norm, cmap=results[0].args.S_cmap),
                cax=ax_cbar, orientation='vertical', label='S', pad = 0.3)
    cbar.set_label("S", rotation=270, labelpad = 10)
    # fig, ax = plt.subplots(figsize=(7,5), dpi = 600)
    trace_cmap = 'viridis'
    num_traces = len(results) -1
    for i, result in enumerate(results):
        if swept_arg == "decay_order":
            # format the decay order as a float with 2 decimal places
            swept_arg_format = ".2f"
            swept_arg_unit = ""
            unformatted_swept_arg_value = result.args.__dict__[swept_arg]
            formatted_swept_arg_value = f"{unformatted_swept_arg_value:{swept_arg_format}}"
            label_string = f"n = {formatted_swept_arg_value}"
            ymin, ymax = -0.5, 0.15
            plt_title = f"sweeping decay order n"
        elif swept_arg == "S_0":
            # format the S_0 as a float with 2 decimal places
            swept_arg_format = ".2f"
            swept_arg_unit = ""
            unformatted_swept_arg_value = result.args.__dict__[swept_arg]
            formatted_swept_arg_value = f"{unformatted_swept_arg_value:{swept_arg_format}}"
            label_string = f"$S_0$ = {formatted_swept_arg_value}"
            ymin, ymax = -0.4, 0.15
            plt_title = f"sweeping orientation near particle $S_0$"
        else:
            # format the thickness as a float with 1 decimal place
            swept_arg_format = ".1f"
            swept_arg_unit = "nm"
            unformatted_swept_arg_value = result.args.__dict__[swept_arg]
            formatted_swept_arg_value = f"{unformatted_swept_arg_value:{swept_arg_format}}"
            label_string = f"{swept_arg} = {formatted_swept_arg_value} {swept_arg_unit}"
            ymin, ymax = -0.45, 0.15
            plt_title = f"sweeping thickness t (nm) of oriented corona"
        result.A.sel(energy=284.7, method = 'nearest').plot(ax = ax1, lw = 1, color = colormaps[trace_cmap](i/num_traces),label = label_string)
        result.A.sel(energy=285.2, method = 'nearest').plot(ax = ax1, lw = 1, color = colormaps[trace_cmap](i/num_traces))
        ax = fig.add_subplot(gs_viz[i // viz_col, i % viz_col: i % viz_col+ 1])
        ax.imshow(result.viz.img_dict["mat_1_S"], origin="lower")
        ax.set_axis_off()
        ax.text(0.5, 0.1, f"{formatted_swept_arg_value} {swept_arg_unit}", horizontalalignment='center', verticalalignment='center', transform=ax.transAxes, color = "white")
    txt = ax1.annotate("284.7 eV", (0.18, 0.05),xycoords = "data", horizontalalignment='center', verticalalignment='center', transform=ax.transAxes, color = "black")
    txt.set_bbox(dict(facecolor='white', alpha=0.7, edgecolor='none'))
    txt = ax1.annotate("285.2 eV", (0.18, -0.05), xycoords = "data", horizontalalignment='center', verticalalignment='center', transform=ax.transAxes, color = "black")
    txt.set_bbox(dict(facecolor='white', alpha=0.7, edgecolor='none'))
    ax1.tick_params(axis = 'y',which = 'major',  direction = 'in')
    ax1.tick_params(axis = 'y',which = 'minor',  direction = 'in')
    ax1.tick_params(axis = 'x',which = 'major',  direction = 'in')
    ax1.tick_params(axis = 'x',which = 'minor',  direction = 'in')
    
    ax1.set_title(plt_title)
    ax1.hlines(0,0.05,0.4,linestyle = 'dotted',color = 'black', zorder = -100, lw = 1)
    ax1.set_xlim(0.05,0.4)
    ax1.set_ylim(ymin,ymax)
    ax1.xaxis.set_major_locator(ticker.AutoLocator())
    ax1.xaxis.set_minor_locator(ticker.AutoMinorLocator())
    ax1.yaxis.set_major_locator(ticker.AutoLocator())
    ax1.yaxis.set_minor_locator(ticker.AutoMinorLocator())

    # ax1.plot(expt[0],-expt[1],lw = 1, color = 'salmon',linestyle = 'dotted',label = '284.7 eV expt')
    # ax1.plot(expt[0],-expt[2],lw = 1, color = 'dodgerblue',linestyle = 'dotted',label = '285.2 eV expt')
    # ax1.set_title("")
    ax1.set_ylabel("Anisotropy (a.u.)")
    ax1.set_xlabel("q (nm$^{-1})$")
    ax1.legend(loc = "center left")
    plt.savefig(f"{swept_arg}_sweep.pdf", bbox_inches='tight', pad_inches=0)
    plt.show()
    plt.close()



import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib import rc, gridspec
import matplotlib
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import numpy as np
import h5py
import datetime
import warnings



def check_NumMat(f, morphology_type):
    morphology_num = f['Morphology_Parameters/NumMaterial'][()]
    
    if morphology_type == 0:
        num_mat = 0
        while f'Euler_Angles/Mat_{num_mat + 1}_Vfrac' in f.keys():
            num_mat +=1
    elif morphology_type == 1:
        num_mat = 0
        while f'Vector_Morphology/Mat_{num_mat + 1}_unaligned' in f.keys():
            num_mat +=1
    
    assert morphology_num==num_mat, 'Number of materials does not match manual count of materials. Recheck hdf5'
    
    return num_mat
    
    

def checkH5_vector(filename):
    # read in vector morphology
    with h5py.File(filename,'r') as f:
        num_mat = check_NumMat(f,morphology_type=1)
        
        ds = f['Vector_Morphology/Mat_1_unaligned'][()]
        
        unaligned = np.zeros((num_mat,*ds.shape))
        alignment = np.zeros((num_mat,*ds.shape,3))
        
        for i in range(0, num_mat):
            unaligned[i,...] = f[f'Vector_Morphology/Mat_{i+1}_unaligned'][()]
            alignment[i,...] = f[f'Vector_Morphology/Mat_{i+1}_alignment'][()]
    
    # assert that the entire morphology has total material equal to 1
    total_material = np.sum(unaligned,axis=0) + np.sum(alignment**2,axis=(0,-1))
    assert np.allclose(total_material,1), 'Not all voxels in morphology have Total Material equal to 1'
    
    # convert vector to Euler for visualization purposes
    S = np.zeros((num_mat, *ds.shape))
    Vfrac = S.copy()
    theta = S.copy()
    psi = S.copy()
    
    S = np.sum(alignment**2,axis=-1)
    Vfrac = unaligned + S
    
    #calculate theta and psi from vectors
    z = np.array([0,0,1])
    x = np.array([1,0,0])
    with np.errstate(invalid='ignore'):
        normed_vectors = alignment/np.sqrt(S[:,:,:,:,np.newaxis])
    np.nan_to_num(normed_vectors,copy=False)
    theta = np.arccos(np.dot(normed_vectors,z))
    psi = np.arccos(np.dot(normed_vectors,x))
    
    return Vfrac, S, theta, psi


def checkH5_euler(filename):
    
    # read in Euler morphology
    with h5py.File(filename,'r') as f:
        num_mat = check_NumMat(f,morphology_type=0)
        
        ds = f['Euler_Angles/Mat_1_Vfrac'][()]
        
        Vfrac = np.zeros((num_mat,*ds.shape))
        S = Vfrac.copy()
        theta = Vfrac.copy()
        psi = Vfrac.copy()

        #'Mat_1_Psi', 'Mat_1_S', 'Mat_1_Theta', 'Mat_1_Vfrac'

        for i in range(0, num_mat):
            Vfrac[i,:,:,:] = f[f'Euler_Angles/Mat_{i+1}_Vfrac']
            S[i,:,:,:] = f[f'Euler_Angles/Mat_{i+1}_S']
            theta[i,:,:,:] = f[f'Euler_Angles/Mat_{i+1}_Theta']
            psi[i,:,:,:] = f[f'Euler_Angles/Mat_{i+1}_Psi']
            
    # wrap psi around from 0 to pi
    psi = np.mod(psi, np.pi)
    
    # assert that the entire morphology has total material equal to 1
    total_material = np.sum(Vfrac,axis=0)
    assert np.allclose(total_material,1), 'Not all voxels in morphology have Total Material equal to 1' 
       
    return Vfrac, S, theta, psi

def checkH5(filename='perp82.hd5',z_slice = 0, subsample = None, outputmat = None, runquiet = False, plotstyle='light'):
    
    #Can plot with light or dark background
    style_dict = {'dark':'dark_background',
                  'light':'default'}
    
    begin_time = datetime.datetime.now()
    
    with h5py.File(filename, 'r') as f:
        
        # check PhysSize is a float
        PhysSize = f['Morphology_Parameters/PhysSize'][()]
        if type(PhysSize) != 'float':
            warnings.warn(f'PhysSize is currently {array.dtype.name} and incompatible with CyRSoXS. Type must be float')
            
        # check morphology type
        if 'Euler_Angles' in f.keys():
            morphology_type = 0
        elif 'Vector_Morphology' in f.keys():
            morphology_type = 1
        else:
            warnings.warn('Neither \"Euler_Angles\" or \"Vector_Morphology\" group detected in hdf5')
            morphology_type = None
    
    if morphology_type == 0:
        Vfrac, S, theta, psi = checkH5_euler(filename)
        num_mat, zdim, ydim, xdim = Vfrac.shape
    elif morphology_type == 1:
        Vfrac, S, theta, psi = checkH5_vector(filename)
        num_mat, zdim, ydim, xdim = Vfrac.shape
        
    for array in [Vfrac, S, theta, psi]:
        if not array.dtype.name.startwith('float')
            warnings.warn(f'Array dtype is currently {array.dtype.name} and incompatible with CyRSoXS. Array dtype must be float')
    
    # test S arrays are between 0 and 1
    for i in range(S.shape[0]):
        assert np.less_equal(S[i,...],1), f'Not all voxels in Mat_{i+1}_S less than or equal to 1'
        assert np.greater_equal(S[i,...],0), f'Negative values are present in Mat_{i+1}_S'
    
    # test for large values in Euler matrices
    if np.any(np.greater(theta,15)):
        warnings.warn('You have large values in your Theta matrix. Make sure you are in radians, not degrees')
    
    if np.any(np.greater(psi,15)):
        warnings.warn('You have large values in your Psi matrix. Make sure you are in radians, not degrees')
    
    if subsample is None:
        subsample = ydim # y dimension
        
    if not runquiet:
        print(f'Dataset dimensions (Z, Y, X): {zdim} × {ydim} × {xdim}')
        print(f'Number of Materials: {num_mat}')
        print('')

    if z_slice > (zdim-1):
        print(f'Error: z_slice of {z_slice} is greater than the maximum index of {zdim-1}. Using z_slice = {zdim-1} instead.')
        z_slice = zdim-1

        
    if not runquiet:
        
        plt.style.use(style_dict[plotstyle])
        font = {'family' : 'sans-serif',
                'sans-serif' : 'DejaVu Sans',
                'weight' : 'regular',
                'size'   : 18}

        rc('font', **font)
        
        for i in range(0,num_mat):    
            print(f'Material {i+1} Vfrac. Min: {np.amin(Vfrac[i,:,:,:])} Max: {np.amax(Vfrac[i,:,:,:])}')
            print(f'Material {i+1} S. Min: {np.amin(S[i,:,:,:])} Max: {np.amax(S[i,:,:,:])}')
            print(f'Material {i+1} theta. Min: {np.amin(theta[i,:,:,:])} Max: {np.amax(theta[i,:,:,:])}')
            print(f'Material {i+1} psi. Min: {np.amin(psi[i,:,:,:])} Max: {np.amax(psi[i,:,:,:])}')
            
          

            fig = plt.figure(figsize=(22, 33))

            gs = gridspec.GridSpec(nrows=5, 
                                   ncols=2, 
                                   figure=fig,
                                   width_ratios= [1,1],
                                   height_ratios=[3, 1, 0.1, 3, 1],
                                   wspace=0.3,
                                   hspace=0.2)


            start = int(ydim/2)-int(subsample/2)
            end = int(ydim/2)+int(subsample/2)

            ax1 = plt.subplot(gs[0, 0])
            Vfracplot = ax1.imshow(Vfrac[i, z_slice, start:end, start:end], cmap = plt.get_cmap('winter'), origin='lower',interpolation='none')
            ax1.set_ylabel('Y index')
            ax1.set_xlabel('X index')
            

            ax1.set_title(f'Material {i+1} Vfrac')
            Vfrac_cbar = plt.colorbar(Vfracplot, fraction=0.040)
            Vfrac_cbar.set_label('Vfrac: volume fraction', rotation=270, labelpad = 22)

            ax2 = plt.subplot(gs[0, 1])
            Splot = ax2.imshow(S[i, z_slice, start:end, start:end], cmap = plt.get_cmap('nipy_spectral'), origin='lower',interpolation='none')
            ax2.set_ylabel('Y index')
            ax2.set_xlabel('X index')
            ax2.set_title(f'Material {i+1} S')
            S_cbar = plt.colorbar(Splot, fraction=0.040)
            S_cbar.set_label('S: orientational order parameter', rotation=270, labelpad = 22)

            ax3 = plt.subplot(gs[1, 0])
            ax3.hist(Vfrac[i,:,:,:].flatten())
            ax3.set_title(f'Material {i+1} Vfrac')
            ax3.set_xlim(left=0)
            ax3.set_xlabel('Vfrac: volume fraction')
            ax3.set_ylabel('num voxels')
            ax3.set_yscale('log')

            ax4 = plt.subplot(gs[1, 1])
            ax4.hist(S[i,:,:,:].flatten())
            ax4.set_title(f'Material {i+1} S')
            ax4.set_xlim(left=0)
            ax4.set_xlabel('S: orientational order parameter')
            ax4.set_ylabel('num voxels')
            ax4.set_yscale('log')

            ax5 = plt.subplot(gs[3, 0])
            norm=matplotlib.colors.Normalize(vmin=0, vmax=np.pi, clip=False)
            thetaplot = ax5.imshow(np.ma.masked_array(theta[i, z_slice, start:end, start:end], np.logical_or(Vfrac[i, z_slice, start:end, start:end]<0.01, S[i, z_slice, start:end, start:end]<0.01)), 
                                   cmap = plt.get_cmap('jet'), norm=norm, origin='lower',interpolation='none')
            ax5.set_ylabel('Y index')
            ax5.set_xlabel('X index')            
            ax5.set_title(f'Material {i+1} theta')
            theta_cbar = plt.colorbar(thetaplot, fraction=0.040)
            theta_cbar.set_label('theta in radians', rotation=270, labelpad = 22)

            ax6 = plt.subplot(gs[3, 1])
            norm=matplotlib.colors.Normalize(vmin=0, vmax=np.pi, clip=False)
            psiplot = ax6.imshow(np.ma.masked_array(psi[i, z_slice, start:end, start:end], np.logical_or(Vfrac[i, z_slice, start:end, start:end]<0.01, S[i, z_slice, start:end, start:end]<0.01)), 
                                 cmap = plt.get_cmap('hsv'), norm=norm, origin='lower',interpolation='none')
            ax6.set_ylabel('Y index')
            ax6.set_xlabel('X index')            
            ax6.set_title(f'Material {i+1} psi')
            psi_cbar = plt.colorbar(psiplot, fraction=0.040)
            psi_cbar.set_label('psi in radians', labelpad = 22)       

            ax7 = plt.subplot(gs[4, 0])
            ax7.hist(theta[i,:,:,:].flatten())
            ax7.set_title(f'Material {i+1} theta')
            ax7.set_xlim(left=0)
            ax7.set_xlabel('theta in radians')
            ax7.set_ylabel('num voxels')
            ax7.set_yscale('log')

            ax8 = plt.subplot(gs[4, 1])
            ax8.hist(psi[i,:,:,:].flatten())
            ax8.set_title(f'Material {i+1} psi')
            ax8.set_xlim(left=0)
            ax8.set_xlabel('psi in radians')
            ax8.set_ylabel('num voxels')
            ax8.set_yscale('log')

            ax4i = inset_axes(ax6,  axes_class = matplotlib.projections.get_projection_class('polar'), width = 2, height = 2, axes_kwargs={"alpha": 0})

            ax4i.grid(False)
            #creates inset legend for orientation
            azimuths = np.deg2rad(np.arange(0, 360, 1))
            zeniths = np.linspace(5,10,50)
            values = np.mod(azimuths, np.pi) * np.ones((50, 360))
            ax4i.plot(0,0, 'o', ms= 75, mec='none', mfc=plt.rcParams['axes.facecolor'], mew=2, zorder = 0, alpha = 0.7)
            ax4i.pcolormesh(azimuths, zeniths, values, cmap=cm.hsv, shading = 'auto')
            ax4i.set_axis_off()
            ax4i.arrow(0,0,0,4,width=0.015,head_width=0.25,head_length=0.5) #,facecolor='k',edgecolor='k')
            ax4i.arrow(np.pi/2,0,0,4,width=0.015,head_width=0.25,head_length=0.5) #,facecolor='k',edgecolor='k')
            ax4i.text(2*np.pi-np.deg2rad(40),3.5,'X')
            ax4i.text(np.pi/2+np.deg2rad(40),3.5,'Y')

            ax5i = inset_axes(ax5,  axes_class = matplotlib.projections.get_projection_class('polar'), width = 2, height = 2, axes_kwargs={"alpha": 0})

            ax5i.grid(False)
            #creates inset legend for orientation
            azimuths_t = np.deg2rad(np.arange(-90, 90, 1))
            zeniths_t = np.linspace(5,10,50)
            values_t = np.mod(np.pi/2-azimuths_t, np.pi) * np.ones((50, 180))
            ax5i.plot(0,0, 'o', ms= 75, mec='none', mfc=plt.rcParams['axes.facecolor'], mew=2, zorder = 0, alpha = 0.7)
            ax5i.pcolormesh(azimuths_t, zeniths_t, values_t, cmap=cm.jet, shading = 'auto')
            ax5i.set_axis_off()
            ax5i.arrow(0,0,0,4,width=0.015,head_width=0.25,head_length=0.5) #,facecolor='k',edgecolor='k')
            ax5i.arrow(np.pi/2,0,0,4,width=0.015,head_width=0.25,head_length=0.5) #,facecolor='k',edgecolor='k')
            ax5i.text(2*np.pi-np.deg2rad(40),3.5,'X')
            ax5i.text(np.pi/2+np.deg2rad(40),3.5,'Z')

            plt.show()
            
    plt.rcParams.update(plt.rcParamsDefault)
    
    # print total run time
    if not runquiet:
        print(f'Total Vfrac whole model. Min: {np.amin(np.sum(Vfrac, axis=0))} Max: {np.amax(np.sum(Vfrac, axis=0))}')
        print(datetime.datetime.now() - begin_time)
    
    # optionally return morphology arrays
    if not (outputmat is None):
        start = int(ydim/2)-int(subsample/2)
        end = int(ydim/2)+int(subsample/2)
        return Vfrac[outputmat,z_slice,start:end,start:end], S[outputmat,z_slice,start:end,start:end], psi[outputmat,z_slice,start:end,start:end], theta[outputmat,z_slice,start:end,start:end]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "NRSS"
author = "Peter Dudenas"
release = "0.1.0"
copyright = (
    ": Official Contribution of the US Government.  Not subject to copyright in the United States"
)

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

import os
import sys

sys.path.insert(0, os.path.abspath("../../"))
sys.path.insert(0, os.path.abspath("../NRSS/"))


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_design",
]  # numpydoc and google docstrings

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "pydata_sphinx_theme"
html_logo = "_static/Logo_NRSS.svg"
html_theme_options = {
    "logo": {
        "image_light": "_static/Logo_NRSS.svg",
        "image_dark": "_static/Logo_NRSS.svg",
    },
    "github_url": "https://github.com/usnistgov/NRSS",
    "collapse_navigation": True,
    "external_links": [
        {
            "name": "NIST RSoXS project",
            "url": "https://www.nist.gov/programs-projects/resonant-soft-x-ray-scattering-rsoxs",
        },
        {
            "name": "PyHyperScattering",
            "url": "https://pages.nist.gov/PyHyperScattering/en/main/index.html",
        },
    ],
    "header_links_before_dropdown": 6,
    # Add light/dark mode and documentation version switcher:
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



# """
# NRSS

# The NIST RSoXS Simulation Suite: A collection of Python and C++/CUDA tools for simulating and analyzing Resonant Soft X-ray Scattering (RSoXS).
# """

# from NRSS import writer
# from NRSS import checkH5
# from NRSS import morphology
# from NRSS import reader

# __version__ = "0.1.0"
# __author__ = "Peter Dudenas"
# __credits__ = "National Institute of Standards and Technology"

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0"  # Fallback version if the file is missing



import numpy as np
import h5py
import datetime
import warnings
import pathlib


def write_hdf5(material_list, PhysSize, fname, MorphologyType=0, ordering='ZYX', author='NIST'):
    '''
    Writes Euler or Vector Morphology format into CyRSoXS-compatible HDF5 file and returns the hdf5 filename

    Parameters
    ----------
        material_list : lists
            List of material lists. 
            Euler Ex. [[Mat_1_Vfrac, Mat_1_S, Mat_1_Theta, Mat_1_Psi],[Mat_2_Vfrac, Mat_2_S, Mat_2_Theta, Mat_2_Psi]]
            Vector Ex. [[Mat_1_alignment, Mat_1_unaligned],[Mat_2_alignment, Mat_2_unaligned]]
        PhysSize : float
            Voxel size
        fname : str or path
            name of hdf5 file to write
        MorphologyType : int
            0 - Euler
            1 - Vector
        ordering : str
            String denoting the axes ordering. 'ZYX' or 'XYZ'
        author : str
            Name of author writing the morphology
    
    Returns
    -------
        fname
            Name of hdf5 file written
    '''

    print(f'--> Marking {fname}')
    with h5py.File(fname, 'w') as f:
        num_mat = len(material_list)
        if MorphologyType == 0:
            i = 1
            for material in material_list:
                f.create_dataset(f"Euler_Angles/Mat_{i}_Vfrac", data=material[0], compression="gzip", compression_opts=9)
                f.create_dataset(f"Euler_Angles/Mat_{i}_S", data=material[1], compression="gzip", compression_opts=9)
                f.create_dataset(f"Euler_Angles/Mat_{i}_Theta", data=material[2], compression="gzip", compression_opts=9)
                f.create_dataset(f"Euler_Angles/Mat_{i}_Psi", data=material[3], compression="gzip", compression_opts=9)
                for j in range(3):
                    f[f"Euler_Angles/Mat_{i}_Vfrac"].dims[j].label = ordering[j]
                    f[f"Euler_Angles/Mat_{i}_S"].dims[j].label = ordering[j]
                    f[f"Euler_Angles/Mat_{i}_Theta"].dims[j].label = ordering[j]
                    f[f"Euler_Angles/Mat_{i}_Psi"].dims[j].label = ordering[j]
                i += 1
        
        elif MorphologyType == 1:
            i = 1
            for material in material_list:
                f.create_dataset(f"Vector_Morphology/Mat_{i}_alignment", data=material[0], compression="gzip", compression_opts=9)
                f.create_dataset(f"Vector_Morphology/Mat_{i}_unalignment", data=material[1], compression="gzip", compression_opts=9)
                for j in range(3):
                    f[f"Vector_Morphology/Mat_{i}_Vfrac"].dims[j].label = ordering[j]
                    f[f"Vector_Morphology/Mat_{i}_S"].dims[j].label = ordering[j]
                    f[f"Vector_Morphology/Mat_{i}_Theta"].dims[j].label = ordering[j]
                    f[f"Vector_Morphology/Mat_{i}_Psi"].dims[j].label = ordering[j]
                i += 1
        else:
            warnings.warn('Check MorphologyType. Should be 0 (Euler) or 1 (Vector)', stacklevel=2)

        f.create_dataset('Morphology_Parameters/creation_date', data=datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
        f.create_dataset('Morphology_Parameters/morphology_creator', data=author)
        f.create_dataset('Morphology_Parameters/PhysSize', data=PhysSize)
        f.create_dataset('Morphology_Parameters/NumMaterial', data=num_mat)

    return fname


def write_config(Energies, EAngleRotation, CaseType=0, MorphologyType=0,
                 NumThreads=4, AlgorithmType=0, DumpMorphology=False,
                 ScatterApproach=0, WindowingType=0, RotMask=False,
                 EwaldsInterpolation=1):
    """
    Writes config.txt file for CyRSoXS simulations

    Parameters
    ----------
        energies : list
            List of energies to be simulated
        EAngleRotation : list
            List of angle rotations in degrees. Format is [AngleStart, AngleIncrement, AngleStop]
        CaseType : int
            Scattering configuration. 
            0 - Default
            1 - Beam Divergence
            2 - Grazing Incidence
        MorphologyType : int
            0 - Euler Morphology (default)
            1 - Vector Morphology
        NumThreads : int
            Number of OpenMP threads to use. Must be >= number of GPUs
        AlgorithmType : int
            0 - Communication minimizing (default)
            1 - Memory minimizing
        DumpMorphology : bool
            Boolean flag to write morphology to file, as seen by CyRSoXS, after necessary conversions are performed
        ScatterApproach : int
            Flag to explicitly calculate the differential scattering cross-section before the Ewald Sphere projection
            0 - Do not compute
            1 - Compute
        WindowingType : int
            Type of FFT window to apply
            0 - None
            1 - Hanning
        RotMask : bool
            Boolean flag to include values outside valid range for all angles
            False - Writes NaNs for pixels that are not valid at every EAngleRotation
            True - Replaces NaNs with values averaged over valid EAngleRotations
        EwaldInterpolation : int
            Type of interpolation for Ewald Sphere projection
            0 - Nearest Neighbor
            1 - Trilinear

    Returns
    -------
        None
    """
    f = open('config.txt', "w")

    # required
    f.write("CaseType = " + str(CaseType) + ";\n")
    f.write("Energies = " + str(Energies) + ";\n")
    f.write("EAngleRotation = " + str(EAngleRotation) + ";\n")
    f.write("MorphologyType = " + str(MorphologyType) + ";\n")

    # optional, but written by default
    f.write("AlgorithmType = " + str(AlgorithmType) + ";\n")
    f.write("NumThreads = " + str(NumThreads) + ";\n")
    f.write("DumpMorphology = " + str(DumpMorphology) + ";\n")
    f.write("ScatterApproach = " + str(ScatterApproach) + ";\n")
    f.write("WindowingType = " + str(WindowingType) + ";\n")
    f.write("RotMask = " + str(RotMask) + ";\n")
    f.write("EwaldsInterpolation = " + str(EwaldsInterpolation) + ";\n")        
    f.close()


# NIST specific
def write_slurm(hdf5_filename, cyrsoxs_version='latest'):
    """
    Writes job.slurm file for use on the NIST glados and wheatley workstations

    Parameters
    ----------
        filename : str or pathlib object
            Name of morphology HDF5 file to be simulated. Can include path
        cyrsoxs_version : str
            String denoting which version of CyRSoXS to run

    Returns
    -------
        None
    """

    f = open("job.slurm","w")
    f.write('#!/bin/bash -l\n')
    f.write('#SBATCH --job-name=CYRSOXS      # Job name\n')
    f.write('#SBATCH --output=cyrsoxs.%j.out # Stdout (%j expands to jobId)\n')
    f.write('#SBATCH --error=cyrsoxs.%j.out  # Stderr (%j expands to jobId)\n')
    f.write('#SBATCH --time=04:00:00         # walltime\n')
    f.write('#SBATCH --nodes=1               # Number of nodes requested (only one node on glados)\n')
    f.write('#SBATCH --ntasks=1              # Number of tasks(processes) (tasks distributed across nodes)\n')
    f.write('#SBATCH --ntasks-per-node=1     # Tasks per node\n')
    f.write('#SBATCH --cpus-per-task=1       # Threads per task (all cpus will be on same node)\n')
    f.write('#SBATCH --gres=gpu:turing:1     # number and type of GPUS to use (glados only has turing)\n')
    f.write('#SBATCH --gres-flags=enforce-binding\n')
    f.write('#SBATCH --partition=gpu\n')
    f.write('\n')
    f.write('set -e\n')
    f.write('\n')
    f.write('if [ x$SLURM_CPUS_PER_TASK == x ]; then\n')
    f.write('    export OMP_NUM_THREADS=1\n')
    f.write('else\n')
    f.write('    export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n')
    f.write('fi\n')
    f.write('\n')
    f.write('# load necessary modules\n')
    f.write('source ~/.typyEnv CLEAN\n')
    f.write(f'typyEnv --add cyrsoxs/{cyrsoxs_version}\n')
    f.write('\n')
    f.write('## RUN YOUR PROGRAM ##\n')
    f.write('echo "RUNNING ON GPU"${CUDA_VISIBLE_DEVICES}\n')
    f.write(f'srun CyRSoXS {hdf5_filename}')
    f.close()


# Set of functions originally authored by Kumar Saurabh @maksbh 
def find_nearest(array, value):
    """
    Function to find the nearest index 

    Parameters
    ----------

    array : ndarray
    value : value of energy

    Returns
    -------
    idx : Integer
        index location corresponding to the closest location
    """
    idx = (np.abs(array - value)).argmin()
    return idx


def get_interpolated_value(array, value, nearest_id, energy_id):
    """
    Function to get the interpolated value

    Parameters
    ----------

    array : ndarray
    value : value of energy
    nearest_id : id corresponding to the nearest value

    Returns
    -------
    valArray : ndarray
            array of the interpolated values
    """
    valArray = np.zeros(array.shape[1])
    if (array[nearest_id][energy_id] > value):
        xp = [array[nearest_id - 1][energy_id], array[nearest_id][energy_id]]
        for i in range(0, array.shape[1]):
            yp = [array[nearest_id - 1][i], array[nearest_id][i]]
            valArray[i] = np.interp(value, xp, yp)

    elif (array[nearest_id][energy_id] < value):
        xp = [array[nearest_id][energy_id], array[nearest_id + 1][energy_id]]
        for i in range(0, array.shape[1]):
            yp = [array[nearest_id][i], array[nearest_id + 1][i]]
            valArray[i] = np.interp(value, xp, yp)

    else:
        for i in range(0, len(valArray)):
            valArray[i] = array[nearest_id][i]

    return valArray


def removeDuplicates(Data, energy_id):
    """
    Function to remove duplicate energies

    Parameters
    ----------

    Data : ndarray
    energy_id : int

    Returns
    -------

    listOut : ndarray

    """
    listIn = Data.tolist()
    listOut = []
    listOut.append(listIn[0])
    currEnergy = listIn[0][energy_id]
    duplicateFound = False
    for i in range(1, len(listIn)):
        if (listIn[i][energy_id] == currEnergy):
            duplicateFound = True
            continue
        else:
            listOut.append(listIn[i])
            currEnergy = listIn[i][energy_id]

    if (duplicateFound):
        print('Duplicates in Energy found. Removing it')
    return (np.array(listOut))


def dump_dataVacuum(index, energy, f):
    """
    Function to write vacuum optical contants (n = 1 + i0)

    Parameters
    ----------

    index : int
    energy : float

    """
    Header = "EnergyData" + str(index) + ":\n{\n"
    f.write(Header)
    Energy = "Energy = " + str(energy) + ";\n"
    f.write(Energy)
    BetaPara = "BetaPara = " + str(0.0) + ";\n"
    f.write(BetaPara)
    BetaPerp = "BetaPerp = " + str(0.0) + ";\n"
    f.write(BetaPerp)
    DeltaPara = "DeltaPara = " + str(0.0) + ";\n"
    f.write(DeltaPara)
    DeltaPerp = "DeltaPerp = " + str(0.0) + ";\n"
    f.write(DeltaPerp)
    f.write("}\n")


def dump_data(valArray, index, labelEnergy, f):
    """
    Function to write material optical constants to file

    Parameters
    ----------

    valArray : ndarray
    index : int
    labelEnergy : dict
    f : file object


    """
    Header = "EnergyData" + str(index) + ":\n{\n";
    f.write(Header)
    Energy = "Energy = " + str(valArray[labelEnergy["Energy"]]) + ";\n"
    f.write(Energy)
    BetaPara = "BetaPara = " + str(valArray[labelEnergy["BetaPara"]]) + ";\n"
    f.write(BetaPara)
    BetaPerp = "BetaPerp = " + str(valArray[labelEnergy["BetaPerp"]]) + ";\n"
    f.write(BetaPerp)
    DeltaPara = "DeltaPara = " + str(valArray[labelEnergy["DeltaPara"]]) + ";\n"
    f.write(DeltaPara)
    DeltaPerp = "DeltaPerp = " + str(valArray[labelEnergy["DeltaPerp"]]) + ";\n"
    f.write(DeltaPerp)
    f.write("}\n")


def writeList(name: str, value: list, file):
    """
    Function to write list to file

    Parameters
    ----------

    name : str
    value : list
    file : file object

    """
    valStr: str = name + "["
    for i in range(len(value) - 1):
        valStr = valStr + str(value[i]) + ","
    valStr = valStr + str(value[len(value) - 1])
    file.write(valStr + "];\n")


def write_materials(energies, materialdict, labelEnergy, numMaterial):
    """
    Function to write optical constants for all energies supplied

    Parameters
    ----------

    energies : ndarray
    materialdict : dict
    labelEnergy : dict
    numMaterial : int


    """
    NumEnergy = len(energies)

    for numMat in range(1, numMaterial+1):
        f = open("Material" + str(numMat) + ".txt", "w")
        fname = materialdict["Material" + str(numMat)]
        if (fname != 'vacuum'):
            Data = np.loadtxt(fname, skiprows=1)
            Data = Data[Data[:, labelEnergy["Energy"]].argsort()]
            Data = removeDuplicates(Data, labelEnergy["Energy"])
            for i in range(0, NumEnergy):
                currentEnergy = energies[i]
                nearest_id = find_nearest(Data[:, labelEnergy["Energy"]], currentEnergy)
                ValArray = get_interpolated_value(Data, currentEnergy, nearest_id, labelEnergy["Energy"])
                dump_data(ValArray, i, labelEnergy, f)

        else:
            for i in range(0, NumEnergy):
                currentEnergy = energies[i]
                dump_dataVacuum(i, currentEnergy, f)
        f.close()


def write_opts(optical_constants, material_num, path):
    # pybind ordering is deltaPara, betaPara, deltaPerp, betaPerp
    # CLI ordering is betaPara, betaPerp, deltaPara, deltaPerp
    if path:
        fname = pathlib.Path(path, "Material" + str(material_num) + ".txt")
    else:
        fname = "Material" + str(material_num) + ".txt"
    f = open(fname, "w")
    for i, key in enumerate(optical_constants):
        Header = "EnergyData" + str(i) + ":\n{\n"
        f.write(Header)
        Energy = "Energy = " + str(key) + ";\n"
        f.write(Energy)
        BetaPara = "BetaPara = " + str(optical_constants[key][1]) + ";\n"
        f.write(BetaPara)
        BetaPerp = "BetaPerp = " + str(optical_constants[key][3]) + ";\n"
        f.write(BetaPerp)
        DeltaPara = "DeltaPara = " + str(optical_constants[key][0]) + ";\n"
        f.write(DeltaPara)
        DeltaPerp = "DeltaPerp = " + str(optical_constants[key][2]) + ";\n"
        f.write(DeltaPerp)
        f.write("}\n")
    f.close()



from lib.util import center_sphere, write_sphere_hdf5, write_config2D, write_slurm
from lib.generateConstants import *
from lib.reduce import *
import pathlib
import os
import datetime
import subprocess
import shlex
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import xarray as xr
import glob
import sys
import numpy as np


def generate_sphere_test(basePath, return_folderpath=False, slurm_file=0, **slurm_kwargs):
    # default input parameters
    NumXY = 2048
    NumZ = 1
    radius = 50
    PhysSize = 5.0 #nm
    NumMat = 2

    # create projection of 3D sphere
    projection = center_sphere(NumXY, PhysSize, radius)
#     projection = np.reshape(projection,(1,512,512)) #ZYX convention

    # write to file
    filename = write_sphere_hdf5(projection,NumXY, PhysSize, radius)


    # write config and material files
    startEnergy = 270.0
    endEnergy = 310.0
    incrementEnergy = 0.2

    numThreads = 4;        #number of threads for execution
    #Files corresponding to Each material.
    dict={'Material0':'PEOlig2018.txt',
          'Material1':'vacuum'}


    # Label of energy to look for
    labelEnergy={"BetaPara":0,
                 "BetaPerp":1,
                 "DeltaPara":2,
                 "DeltaPerp":3,
                 "Energy":6}

    ens = np.round(np.arange(startEnergy,endEnergy+incrementEnergy,incrementEnergy),2)
    write_config2D(NumXY,PhysSize,ens,euler=True)

    # renamed "main" in generateConstants.py to something more descriptive
    write_materials(startEnergy,endEnergy,incrementEnergy,dict,labelEnergy,len(dict))

    # Create folder stamped with current datetime
    time_now = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    folder_name = 'spheretest_' + time_now
    os.mkdir(folder_name)
    subprocess.run(["mv","Material1.txt",f"./{folder_name}"])
    subprocess.run(["mv","Material2.txt",f"./{folder_name}"])
    subprocess.run(["mv","config.txt",f"./{folder_name}"])
    subprocess.run(["mv",f"{filename}",f"./{folder_name}"])

    if slurm_file == 1:
        write_slurm(filename, **slurm_kwargs)
        subprocess.run(["mv","job.slurm",f"./{folder_name}"])
        
        
    if return_folderpath:
        return pathlib.Path(basePath,folder_name)
        
def run_slurm(folder_path):
    os.chdir(folder_path)
    subprocess.run(['sbatch','job.slurm'])
    

rho = complex(-0.0001569718509412237,0.0001268706792520359) #284.98651 eV
h          = 4.135667696e-15 #eV*s
c          = 299792458 #m/s
wavelength =  h*c/284.98651
SLD = rho*2*np.pi/(wavelength*1e9)**2

def analytical_sphere(q, radius, NumXY, PhysSize):
    vol = 4/3*np.pi*(radius)**3
    scale = vol/(NumXY/PhysSize)**3
    I = scale/vol*np.abs(3*vol*SLD*(np.sin(q*radius)-q*radius*np.cos(q*radius))/(q*radius)**3)**2
    return I
    
def test_vs_analytical(folder_path,savepng=False):
    h5_path = pathlib.Path(folder_path,'HDF5')
    h5s = sorted(list(h5_path.glob('*h5')))
    energy = np.zeros(len(h5s))
    data = read_img_h5(h5s[0], PhysSize=5)
    remeshed_data = remesh(data)
    energy[0] = h5s[0].name.split('_')[1][:-3]
    all_remeshed = np.zeros((len(h5s),remeshed_data.values.shape))
    all_remeshed[0,...] = remeshed_data.values
    for i in range(1,len(h5s)):
        energy[i] = h5s[i].name.split('_')[1][:-3]
        tmp = read_img_h5(h5s[i],PhysSize=5)
        tmp_remesh = remesh(tmp)
        all_remeshed[i,...] = tmp_remesh.values
    
    all_remeshed = xr.DataArray(all_remeshed,dims=['energy','chi','q'],coords={'energy':energy,'chi':tmp_remesh.chi,'q':tmp_remesh.q})
    
    I_simulation = remeshed_data.mean('chi',)
    I_simulation = I_simulation.where((I_simulation.q < 0.6) & (I_simulation.q > 2e-3),drop=True)
    I_analytical = analytical_sphere(I_simulation.q.values,50,2048,5)
#     avg_scaling = np.nanmean(I_analytical/I_simulation)
    I_simulation = I_simulation*I_analytical[0]/I_simulation[0]
    
    if savepng:

        plt.figure()
        plt.loglog(I_simulation.q,I_analytical,label='Analytical')
        I_simulation.plot(yscale='log',xscale='log',label='Simulation',linestyle='--')
        plt.legend()
        plt.xlabel(r'q [nm$^{-1}$]')
        plt.ylabel('I(q)')
        plt.savefig(pathlib.Path(folder_path,'analytical_v_sim.png'),format='png',dpi=140,bbox_inches='tight')
        plt.close()

        plt.figure()
        remeshed_data.plot(norm=LogNorm(1e-10,1e-3))
        plt.title('Remeshed CyRSoXS Output')
        plt.xlabel(r'q [nm$^{-1}$]')
        plt.ylabel(r'$\chi$ [Degrees]')
        plt.savefig(pathlib.Path(folder_path,'remeshed.png'),format='png',dpi=140,bbox_inches='tight')
        plt.close()

        plt.figure()
        data.plot(norm=LogNorm(1e-10,1e-3))
        plt.title('Raw CyRSoXS output')
        plt.xlabel(r'q$_x$ [nm$^{-1}$]')
        plt.ylabel(r'q$_y$ [nm$^{-1}$]')
        plt.savefig(pathlib.Path(folder_path,'raw.png'),format='png',dpi=140,bbox_inches='tight')
        plt.close()
        
        plt.figure()
        plt.loglog(I_simulation.q, np.abs(I_simulation - I_analytical))
        plt.title(r'(I$_{simulation}$ - I$_{analytical}$)')
        plt.xlabel(r'q [nm$^{-1}$]')        
        plt.ylabel('I(q)')
        plt.savefig(pathlib.Path(folder_path,'difference.png'),format='png',dpi=140,bbox_inches='tight')
        plt.close()
        
        plt.figure()
        all_remeshed.mean('chi').sum('q').plot()
        plt.xlabel('Energy [eV]')
        plt.ylabel('ISI')
        plt.savefig(pathlib.path(folder_path,'ISI.png'),format='png',dpi=140,bbox_inches='tight')
        plt.close()
        
        
    
#     mean_diff = np.abs(np.mean(I_simulation - I_analytical))
    assert(np.allclose(I_simulation,I_analytical,1e-7))
    
    return I_simulation, I_analytical
    
if __name__ == '__main__':
    version = 'latest'
    slurm_file = 0
    
    for arg in sys.argv:
        if arg.startswith('slurm'):
            slurm_file = int(arg.split('=')[-1])
        if arg.startswith('version'):
            version = arg.split('=')[-1]

    basePath = pathlib.Path('.').absolute()
    
    workpath = generate_sphere_test(basePath,return_folderpath=True, slurm_file=slurm_file,cyrsoxs_version=version)

    if slurm_file == 1:
        import time
        run_slurm(workpath)
        hdf5path = pathlib.Path(workpath,'HDF5')
        while len(list(hdf5path.glob('*.h5'))) < 201:
            time.sleep(10)
#         test_vs_analytical(workpath,savepng=True)
        I_simulation, I_analytical = test_vs_analytical(workpath, savepng=True)